"""add calories to workouts + add user preference for visibility

Revision ID: 267735b7574f
Revises: ef18bfbcc613
Create Date: 2025-12-25 17:31:16.887230

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '267735b7574f'
down_revision = 'ef18bfbcc613'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.add_column(sa.Column('calories', sa.Integer(), nullable=True))

    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('calories_visibility', sa.Enum('PUBLIC', 'FOLLOWERS', 'PRIVATE', name='visibility_levels'), server_default='PRIVATE', nullable=False))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.drop_column('calories_visibility')

    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.drop_column('calories')

    # ### end Alembic commands ###



"""add weather infos in 'Activity' table

Revision ID: 71093ac9ca44
Revises: e82e5e9447de
Create Date: 2018-06-13 15:29:11.715377

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '71093ac9ca44'
down_revision = 'e82e5e9447de'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('activities', sa.Column('weather_end', sa.JSON(), nullable=True))
    op.add_column('activities', sa.Column('weather_start', sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('activities', 'weather_start')
    op.drop_column('activities', 'weather_end')
    # ### end Alembic commands ###



from typing import Dict, Union

from flask import Blueprint, request
from sqlalchemy import exc

from fittrackee import db
from fittrackee.oauth2.server import require_auth
from fittrackee.responses import (
    DataNotFoundErrorResponse,
    HttpResponse,
    InvalidPayloadErrorResponse,
    handle_error_and_return_response,
)
from fittrackee.users.models import User, UserSportPreference
from fittrackee.users.roles import UserRole

from .models import Sport

sports_blueprint = Blueprint("sports", __name__)


@sports_blueprint.route("/sports", methods=["GET"])
@require_auth(
    scopes=["workouts:read"],
    optional_auth_user=True,
    allow_suspended_user=True,
)
def get_sports(auth_user: User) -> Dict:
    """
    Get all sports.

    For Hiking, Running, Trail and Walking, ``pace_speed_display`` is returned.

    Suspended user can access this endpoint.

    **Scope**: ``workouts:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/sports HTTP/1.1
      Content-Type: application/json

    **Example response**:

    - for non admin user:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "sports": [
            {
              "color": null,
              "id": 1,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Cycling (Sport)",
              "stopped_speed_threshold": 1
            },
            {
              "color": null,
              "id": 2,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Cycling (Transport)",
              "stopped_speed_threshold": 1
            },
            {
              "color": null,
              "id": 3,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Hiking",
              "pace_speed_display": "pace",
              "stopped_speed_threshold": 0.1
            },
            {
              "color": null,
              "id": 4,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Mountain Biking",
              "stopped_speed_threshold": 1
            },
            {
              "color": null,
              "id": 5,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Running",
              "pace_speed_display": "pace",
              "stopped_speed_threshold": 0.1
            },
            {
              "color": null,
              "id": 6,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Walking",
              "pace_speed_display": "pace",
              "stopped_speed_threshold": 0.1
            }
          ]
        },
        "status": "success"
      }

    - for admin user and check_workouts=true:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "sports": [
            {
              "color": null,
              "has_workouts": true,
              "id": 1,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Cycling (Sport)",
              "stopped_speed_threshold": 1
            },
            {
              "color": null,
              "has_workouts": false,
              "id": 2,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Cycling (Transport)",
              "stopped_speed_threshold": 1
            },
            {
              "color": null,
              "has_workouts": false,
              "id": 3,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Hiking",
              "pace_speed_display": "pace",
              "stopped_speed_threshold": 0.1
            },
            {
              "color": null,
              "has_workouts": false,
              "id": 4,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Mountain Biking",
              "stopped_speed_threshold": 1
            },
            {
              "color": null,
              "has_workouts": false,
              "id": 5,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Running",
              "pace_speed_display": "pace",
              "stopped_speed_threshold": 0.1
            },
            {
              "color": null,
              "has_workouts": false,
              "id": 6,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Walking",
              "pace_speed_display": "pace",
              "stopped_speed_threshold": 0.1
            }
          ]
        },
        "status": "success"
      }

    :query boolean check_workouts: check if sport has workouts

    :reqheader Authorization: OAuth 2.0 Bearer Token if user is authenticated

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``

    """
    params = request.args.copy()
    check_workouts = params.get("check_workouts", "false").lower() == "true"

    sport_preferences = (
        {
            p.sport_id: p
            for p in UserSportPreference.query.filter_by(
                user_id=auth_user.id
            ).all()
        }
        if auth_user
        else {}
    )
    sports = Sport.query.order_by(Sport.id).all()
    sports_data = []
    for sport in sports:
        sports_data.append(
            sport.serialize(
                check_workouts=(
                    auth_user and auth_user.has_admin_rights and check_workouts
                ),
                sport_preferences=(
                    sport_preferences[sport.id].serialize()
                    if sport.id in sport_preferences
                    else None
                ),
            )
        )
    return {
        "status": "success",
        "data": {"sports": sports_data},
    }


@sports_blueprint.route("/sports/<int:sport_id>", methods=["GET"])
@require_auth(scopes=["workouts:read"])
def get_sport(auth_user: User, sport_id: int) -> Union[Dict, HttpResponse]:
    """
    Get a sport.

    For Hiking, Running, Trail and Walking, ``pace_speed_display`` is returned.

    **Scope**: ``workouts:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/sports/1 HTTP/1.1
      Content-Type: application/json

    **Example response**:

    - success for non admin user:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "sports": [
            {
              "color": null,
              "id": 1,
              "is_active": true,
              "is_active_for_user": true,
              "label": "Cycling (Sport)",
              "stopped_speed_threshold": 1
            }
          ]
        },
        "status": "success"
      }

    - sport not found:

    .. sourcecode:: http

      HTTP/1.1 404 NOT FOUND
      Content-Type: application/json

      {
        "data": {
          "sports": []
        },
        "status": "not found"
      }

    :param integer sport_id: sport id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``sport not found``

    """
    sport = Sport.query.filter_by(id=sport_id).first()
    if sport:
        sport_preferences = UserSportPreference.query.filter_by(
            user_id=auth_user.id, sport_id=sport.id
        ).first()
        return {
            "status": "success",
            "data": {
                "sports": [
                    sport.serialize(
                        sport_preferences=(
                            sport_preferences.serialize()
                            if sport_preferences
                            else None
                        ),
                    )
                ]
            },
        }
    return DataNotFoundErrorResponse("sports")


@sports_blueprint.route("/sports/<int:sport_id>", methods=["PATCH"])
@require_auth(scopes=["workouts:write"], role=UserRole.ADMIN)
def update_sport(auth_user: User, sport_id: int) -> Union[Dict, HttpResponse]:
    """
    Update a sport.

    For Hiking, Running, Trail and Walking, ``pace_speed_display`` is returned.

    Authenticated user must be an admin.

    **Scope**: ``workouts:write``

    **Minimum role**: Administrator

    **Example request**:

    .. sourcecode:: http

      PATCH /api/sports/1 HTTP/1.1
      Content-Type: application/json

    **Example responses**:

    - success:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "sports": [
            {
              "color": null,
              "has_workouts": false,
              "id": 1,
              "is_active": false,
              "is_active_for_user": false,
              "label": "Cycling (Sport)",
              "stopped_speed_threshold": 1
            }
          ]
        },
        "status": "success"
      }

    - sport not found:

    .. sourcecode:: http

      HTTP/1.1 404 NOT FOUND
      Content-Type: application/json

      {
        "data": {
          "sports": []
        },
        "status": "not found"
      }

    :param integer sport_id: sport id

    :<json string is_active: sport active status

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: sport updated
    :statuscode 400:
        - ``invalid payload``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``sport not found``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    sport_data = request.get_json()
    if not sport_data or sport_data.get("is_active") is None:
        return InvalidPayloadErrorResponse()

    try:
        sport = Sport.query.filter_by(id=sport_id).first()
        if not sport:
            return DataNotFoundErrorResponse("sports")

        sport.is_active = sport_data.get("is_active")
        db.session.commit()
        sport_preferences = UserSportPreference.query.filter_by(
            user_id=auth_user.id, sport_id=sport.id
        ).first()
        return {
            "status": "success",
            "data": {
                "sports": [
                    sport.serialize(
                        sport_preferences=(
                            sport_preferences.serialize()
                            if sport_preferences
                            else None
                        ),
                    )
                ]
            },
        }

    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)



"""add preference for workout chart display

Revision ID: 98688d3d293b
Revises: 4eeeecd3936d
Create Date: 2025-07-19 18:34:49.412724

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '98688d3d293b'
down_revision = '4eeeecd3936d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('split_workout_charts', sa.Boolean(), server_default='false', nullable=False))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.drop_column('split_workout_charts')

    # ### end Alembic commands ###



from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Dict, List, Optional
from uuid import UUID, uuid4

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.sql.expression import text
from sqlalchemy.types import Enum

from fittrackee import BaseModel, db
from fittrackee.database import TZDateTime
from fittrackee.dates import aware_utc_now
from fittrackee.utils import encode_uuid
from fittrackee.visibility_levels import VisibilityLevel, can_view

if TYPE_CHECKING:
    from fittrackee.users.models import User, UserSportPreference
    from fittrackee.workouts.models import Workout

from .exceptions import EquipmentForbiddenException

WorkoutEquipment = db.Table(
    "workout_equipments",
    db.Column(
        "workout_id",
        db.Integer,
        db.ForeignKey("workouts.id", ondelete="CASCADE"),
        primary_key=True,
    ),
    db.Column(
        "equipment_id",
        db.Integer,
        db.ForeignKey("equipments.id", ondelete="CASCADE"),
        primary_key=True,
        index=True,
    ),
)


class Equipment(BaseModel):
    __tablename__ = "equipments"
    # a single user can only have one equipment with the same label
    __table_args__ = (
        UniqueConstraint(
            "label",
            "user_id",
            name="equipment_user_label_unique",
        ),
    )

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    uuid: Mapped[UUID] = mapped_column(
        postgresql.UUID(as_uuid=True),
        default=uuid4,
        unique=True,
        nullable=False,
    )
    user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id"), index=True, nullable=False
    )
    label: Mapped[str] = mapped_column(
        db.String(50), unique=False, nullable=False
    )
    description: Mapped[Optional[str]] = mapped_column(
        db.String(200), default=None, nullable=True
    )
    equipment_type_id: Mapped[int] = mapped_column(
        db.ForeignKey("equipment_types.id")
    )
    creation_date: Mapped[datetime] = mapped_column(
        TZDateTime, default=aware_utc_now
    )
    is_active: Mapped[bool] = mapped_column(default=True, nullable=False)
    total_distance: Mapped[float] = mapped_column(
        db.Numeric(10, 3),
        nullable=False,
        server_default=text("0.0"),  # kilometers
    )
    total_duration: Mapped[timedelta] = mapped_column(
        db.Interval, nullable=False, server_default=text("'00:00:00'")
    )
    total_moving: Mapped[timedelta] = mapped_column(
        db.Interval, nullable=False, server_default=text("'00:00:00'")
    )
    total_workouts: Mapped[int] = mapped_column(
        nullable=False, server_default=text("0")
    )
    visibility: Mapped[VisibilityLevel] = mapped_column(
        Enum(VisibilityLevel, name="visibility_levels"),
        server_default="PRIVATE",
        nullable=False,
    )

    user: Mapped["User"] = relationship(
        "User", lazy="select", single_parent=True
    )
    workouts: Mapped[List["Workout"]] = relationship(
        "Workout", secondary=WorkoutEquipment, back_populates="equipments"
    )
    equipment_type: Mapped["EquipmentType"] = relationship(
        "EquipmentType", lazy="joined", single_parent=True
    )
    default_for_sports: Mapped[List["UserSportPreference"]] = relationship(
        "UserSportPreference",
        primaryjoin=(
            "Equipment.id == "
            "users_sports_preferences_equipments.c.equipment_id"
        ),
        secondaryjoin="""and_(
            UserSportPreference.user_id == 
            users_sports_preferences_equipments.c.user_id,
            UserSportPreference.sport_id == 
            users_sports_preferences_equipments.c.sport_id,
        )""",
        secondary="users_sports_preferences_equipments",
        lazy="select",
        back_populates="default_equipments",
    )

    def __repr__(self) -> str:
        return f"<Equipment {self.id!r} {self.label!r}>"

    def __init__(
        self,
        label: str,
        equipment_type_id: int,
        description: str,
        user_id: int,
        is_active: bool,
    ) -> None:
        self.label = label
        self.equipment_type_id = equipment_type_id
        self.description = description
        self.user_id = user_id
        self.is_active = is_active

    @property
    def short_id(self) -> str:
        return encode_uuid(self.uuid)

    def get_distance(self) -> Optional[float]:
        if self.equipment_type.label == "Racket":
            return None
        return (
            0.0 if self.total_distance is None else float(self.total_distance)
        )

    def serialize(self, *, current_user: Optional["User"]) -> Dict:
        if not can_view(self, "visibility", current_user):
            raise EquipmentForbiddenException()

        serialized_equipment = {
            "label": self.label,
            "is_active": self.is_active,
            "equipment_type": self.equipment_type.serialize(),
        }

        if not current_user or current_user.id != self.user_id:
            return serialized_equipment

        serialized_equipment = {
            **serialized_equipment,
            "id": self.short_id,
            "user_id": self.user_id,
            "description": self.description,
            "default_for_sport_ids": [
                sport_preference.sport_id
                for sport_preference in self.default_for_sports
            ],
            "creation_date": self.creation_date,
            "total_distance": self.get_distance(),
            "total_duration_in_hours": (
                0
                if self.total_moving is None
                else round(self.total_moving.total_seconds() / 3600, 1)
            ),
            "visibility": self.visibility,
            "workouts_count": (
                0 if self.total_workouts is None else self.total_workouts
            ),
        }
        return serialized_equipment


class EquipmentType(BaseModel):
    __tablename__ = "equipment_types"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    label: Mapped[str] = mapped_column(
        db.String(50), unique=False, nullable=False
    )
    is_active: Mapped[bool] = mapped_column(default=True, nullable=False)

    equipments: Mapped[List["Equipment"]] = relationship(
        "Equipment", lazy="select", back_populates="equipment_type"
    )

    def __repr__(self) -> str:
        return f"<EquipmentType {self.label!r}>"

    def __init__(self, label: str, is_active: bool) -> None:
        self.label = label
        self.is_active = is_active

    def serialize(self, is_admin: bool = False) -> Dict:
        serialized_equipment_type = {
            "id": self.id,
            "label": self.label,
            "is_active": self.is_active,
        }
        if is_admin:
            serialized_equipment_type["has_equipments"] = (
                len(self.equipments) > 0
            )
        return serialized_equipment_type



"""update 'users_sports_preferences_equipments' table and add new equipment types

Revision ID: 076d33eab04f
Revises: ff53544547cd
Create Date: 2026-02-15 16:36:52.694015

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "076d33eab04f"
down_revision = "ff53544547cd"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table(
        "users_sports_preferences_equipments", schema=None
    ) as batch_op:
        batch_op.add_column(
            sa.Column("equipment_type_id", sa.Integer(), nullable=True)
        )
        batch_op.create_foreign_key(
            "users_sports_preferences_equipments_equipment_types_fkey",
            "equipment_types",
            ["equipment_type_id"],
            ["id"],
            ondelete="CASCADE",
        )

    equipments_helper = sa.Table(
        "equipments",
        sa.MetaData(),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("equipment_type_id", sa.Integer(), nullable=False),
    )

    connection = op.get_bind()
    for equipment in connection.execute(equipments_helper.select()):
        op.execute(f"""
            UPDATE users_sports_preferences_equipments
            SET equipment_type_id = {equipment.equipment_type_id}
            WHERE users_sports_preferences_equipments.equipment_id = {equipment.id}
            """)

    op.alter_column(
        "users_sports_preferences_equipments",
        "equipment_type_id",
        nullable=False,
    )

    op.execute(
        """
        INSERT INTO equipment_types (label, is_active)
        VALUES ('Racket', True), ('Paddle', True), ('Misc', True);
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM equipment_types
        WHERE label IN ('Racket', 'Paddle', 'Misc');
        """
    )

    with op.batch_alter_table(
        "users_sports_preferences_equipments", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "users_sports_preferences_equipments_equipment_types_fkey",
            type_="foreignkey",
        )
        batch_op.drop_column("equipment_type_id")



"""add Mountaineering

Revision ID: a8cc0adfe1d3
Revises: bf13b8f5589d
Create Date: 2022-12-11 11:03:01.216734

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'a8cc0adfe1d3'
down_revision = 'bf13b8f5589d'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Mountaineering', True, 0.1)
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Mountaineering';
        """
    )



from datetime import datetime, timezone
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import pytz

from fittrackee.utils import decode_short_id
from fittrackee.visibility_levels import can_view

from ..exceptions import WorkoutForbiddenException
from ..models import Workout

if TYPE_CHECKING:
    from fittrackee.users.models import User


def get_workout_datetime(
    workout_date: Union["datetime", str],
    user_timezone: Optional[str],
    date_str_format: Optional[str] = None,
    with_user_timezone: bool = False,
) -> Tuple["datetime", Optional["datetime"]]:
    """
    Return datetime in UTC and datetime in user timezone if with_user_timezone
    is True.

    Note: columns in Datetime are still stored without timezone.
    Values must be in the UTC timezone before storing in database.
    """
    workout_date_in_user_tz = None

    # workout w/o gpx
    if isinstance(workout_date, str):
        if not date_str_format:
            date_str_format = "%Y-%m-%d %H:%M:%S"
        workout_date = datetime.strptime(workout_date, date_str_format)  # noqa: DTZ007
        if user_timezone:
            workout_date = pytz.timezone(user_timezone).localize(workout_date)

    if workout_date.tzinfo is None:
        workout_date_in_utc = workout_date.replace(tzinfo=timezone.utc)
        if user_timezone and with_user_timezone:
            workout_date_in_user_tz = workout_date_in_utc.astimezone(
                pytz.timezone(user_timezone)
            )
    else:
        workout_date_in_utc = workout_date.astimezone(pytz.utc)
        if user_timezone and with_user_timezone:
            workout_date_in_user_tz = workout_date.astimezone(
                pytz.timezone(user_timezone)
            )

    return workout_date_in_utc, workout_date_in_user_tz


def get_datetime_from_request_args(
    params: Dict, user: "User"
) -> Tuple[Optional["datetime"], Optional["datetime"]]:
    date_from = None
    date_to = None

    date_from_str = params.get("from")
    if date_from_str:
        date_from, _ = get_workout_datetime(
            workout_date=date_from_str,
            user_timezone=user.timezone,
            date_str_format="%Y-%m-%d",
        )
    date_to_str = params.get("to")
    if date_to_str:
        date_to, _ = get_workout_datetime(
            workout_date=f"{date_to_str} 23:59:59",
            user_timezone=user.timezone,
        )
    return date_from, date_to


def get_average_speed(
    total_workouts: int,
    total_average_speed: float,
    workout_average_speed: float,
) -> float:
    return round(
        (
            (total_average_speed * (total_workouts - 1))
            + float(workout_average_speed)
        )
        / total_workouts,
        2,
    )


def get_ordered_workouts(
    workouts: List["Workout"], limit: int
) -> List["Workout"]:
    return sorted(
        workouts, key=lambda workout: workout.workout_date, reverse=True
    )[:limit]


def get_workout(
    workout_short_id: str,
    auth_user: Optional["User"],
    allow_admin: bool = False,
) -> "Workout":
    workout_uuid = decode_short_id(workout_short_id)
    workout = Workout.query.filter(Workout.uuid == workout_uuid).first()
    if not workout or (
        not can_view(workout, "workout_visibility", auth_user)
        and not (allow_admin and auth_user and auth_user.has_admin_rights)
    ):
        raise WorkoutForbiddenException()
    return workout



"""move pace speed display preferences to sports

Revision ID: d8416412506b
Revises: 0bbe0c8372cf
Create Date: 2026-01-01 11:51:21.071755

Note: since the migration will be released in the same version as the previous
migrations introducing pace display preference, updating existing data is not
necessary.
"""

from alembic import op
import sqlalchemy as sa

from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "d8416412506b"
down_revision = "0bbe0c8372cf"
branch_labels = None
depends_on = None

pace_speed_display = postgresql.ENUM(
    "PACE",
    "SPEED",
    "PACE_AND_SPEED",
    name="pace_speed_display",
)


def upgrade():
    pace_speed_display.create(op.get_bind(), checkfirst=True)
    with op.batch_alter_table("sports", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "pace_speed_display",
                pace_speed_display,
                server_default="SPEED",
                nullable=False,
            )
        )

    op.execute(
        """
        UPDATE sports
        SET pace_speed_display = 'PACE'
        WHERE label in (
              'Hiking', 'Running', 'Trail', 'Walking'
        );
        """
    )

    with op.batch_alter_table(
        "users_sports_preferences", schema=None
    ) as batch_op:
        batch_op.add_column(
            sa.Column(
                "pace_speed_display",
                pace_speed_display,
                server_default="SPEED",
                nullable=False,
            )
        )

    op.execute(
        """
        UPDATE users_sports_preferences as usp
        SET pace_speed_display = 'PACE'
        FROM (
            SELECT s.id
            FROM sports s
            WHERE s.label in (
                'Hiking', 'Running', 'Trail', 'Walking'
            )
        ) AS s(id)
        WHERE usp.sport_id = s.id;
        """
    )

    with op.batch_alter_table("users", schema=None) as batch_op:
        batch_op.drop_column("display_speed_with_pace")


def downgrade():
    with op.batch_alter_table("users", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "display_speed_with_pace",
                sa.BOOLEAN(),
                server_default=sa.text("false"),
                autoincrement=False,
                nullable=False,
            )
        )

    with op.batch_alter_table(
        "users_sports_preferences", schema=None
    ) as batch_op:
        batch_op.drop_column("pace_speed_display")

    with op.batch_alter_table("sports", schema=None) as batch_op:
        batch_op.drop_column("pace_speed_display")

    pace_speed_display.drop(op.get_bind())



from typing import Dict

from flask import Blueprint
from sqlalchemy.sql import select
from sqlalchemy.sql import text as sql_text

from fittrackee import db
from fittrackee.oauth2.server import require_auth
from fittrackee.users.models import User

from .constants import SPORTS_WITHOUT_ELEVATION_DATA
from .models import Record

records_blueprint = Blueprint("records", __name__)


@records_blueprint.route("/records", methods=["GET"])
@require_auth(scopes=["workouts:read"])
def get_records(auth_user: User) -> Dict:
    """
    Get all records for authenticated user.

    Following types of records are available, depending en sport:
        - average pace (record_type: ``AP``)
        - average speed (record_type: ``AS``)
        - best pace (record_type: ``BP``)
        - farthest distance (record_type: ``FD``)
        - highest ascent (record_type: ``HA``)
        - longest duration (record_type: ``LD``)
        - maximum speed (record_type: ``MS``)

    **Scope**: ``workouts:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/records HTTP/1.1
      Content-Type: application/json

    **Example responses**:

    - returning records

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "records": [
            {
              "id": 14,
              "record_type": "AP",
              "sport_id": 5,
              "user": "admin",
              "value": "0:07:01",
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 9,
              "record_type": "AS",
              "sport_id": 5,
              "user": "admin",
              "value": 8.55,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 15,
              "record_type": "BP",
              "sport_id": 5,
              "user": "admin",
              "value": "0:05:58",
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 10,
              "record_type": "FD",
              "sport_id": 5,
              "user": "admin",
              "value": 2.858,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 13,
              "record_type": "HA",
              "sport_id": 5,
              "user": "Sam",
              "value": 7.029,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 11,
              "record_type": "LD",
              "sport_id": 5,
              "user": "admin",
              "value": "0:20:24",
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 12,
              "record_type": "MS",
              "sport_id": 5,
              "user": "admin",
              "value": 10.06,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            }
          ]
        },
        "status": "success"
      }

    - no records

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "records": []
        },
        "status": "success"
      }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``

    """
    params = {
        "user_id": auth_user.id,
        "sports_without_elevation_data": tuple(SPORTS_WITHOUT_ELEVATION_DATA),
    }
    sql = """
          SELECT records.*
          FROM records
          JOIN sports ON records.sport_id = sports.id
          LEFT OUTER JOIN users_sports_preferences usp 
                       ON usp.sport_id = records.sport_id AND
                          usp.user_id = records.user_id
          WHERE records.user_id = :user_id AND
            (
              records.record_type IN ('FD', 'LD') OR ( 
                records.record_type = 'HA' AND
                sports.label NOT IN :sports_without_elevation_data
              ) OR (
                records.record_type IN ('AS', 'MS') AND
                (
                  usp.pace_speed_display <> 'PACE' OR ( 
                    usp.pace_speed_display IS NULL AND
                    sports.pace_speed_display <> 'PACE'
                  )
                )
              ) OR (
                records.record_type IN ('AP', 'BP') AND
                (
                  usp.pace_speed_display <> 'SPEED' OR (
                    usp.pace_speed_display IS NULL AND
                    sports.pace_speed_display <> 'SPEED'
                  )
                )
              )
            )
          ORDER BY records.sport_id, records.record_type;"""

    records = db.session.scalars(  # type: ignore
        select(Record).from_statement(sql_text(sql)).params(**params)
    ).all()

    return {
        "status": "success",
        "data": {"records": [record.serialize() for record in records]},
    }



from typing import TYPE_CHECKING, Dict, Union

from flask import Blueprint, request

from fittrackee import appLog
from fittrackee.oauth2.server import require_auth

from ..responses import InternalServerErrorResponse
from .nominatim_service import NominatimService

if TYPE_CHECKING:
    from fittrackee.users.models import User

    from ..responses import HttpResponse


geocode_blueprint = Blueprint("geocode", __name__)

nominatim_service = NominatimService()


@geocode_blueprint.route("/geocode/search", methods=["GET"])
@require_auth(scopes=["geocode:read"])
def get_coordinates_from_city(
    auth_user: "User",
) -> Union[Dict, "HttpResponse"]:
    """
    Return coordinates based on location (city) using Nominatim API.

    **Scope**: ``geocode:read``

    **Example requests**:

    .. sourcecode:: http

      GET /api/geocode/search?city=Paris HTTP/1.1

    **Example responses**:

    - returning at least one location:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "locations": [
              {
                "addresstype": "suburb",
                "coordinates": "48.8588897,2.3200410",
                "display_name": "Paris, Île-de-France, France métropolitaine",
                "name": "Paris",
                "osm_id": "r7444",
              }
            ]
          },
          "status": "success"
        }

    - returning no locations

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
            "data": {
                "locations": []
            },
            "status": "success"
        }

    :query string city: location
    :query string language: preferred language for Nominatim results

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 500:
        - ``error when getting coordinates from location``

    """
    params = request.args.copy()
    city = params.get("city")
    language = params.get("language")

    locations = []
    if city:
        try:
            # search is case-insensitive
            locations = nominatim_service.get_locations_from_city(
                city.lower(), language
            )
        except Exception:
            message = "error when getting coordinates from location"
            appLog.exception(message)
            return InternalServerErrorResponse(message)

    return {"status": "success", "locations": locations}


@geocode_blueprint.route("/geocode/lookup", methods=["GET"])
@require_auth(scopes=["geocode:read"])
def get_location_from_id(auth_user: "User") -> Union[Dict, "HttpResponse"]:
    """
    Return location based on OSM id using Nominatim API.

    **Scope**: ``geocode:read``

    **Example requests**:

    .. sourcecode:: http

      GET /api/geocode/lookup?osm_id=r71525 HTTP/1.1

    **Example responses**:

    - returning location:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "location": {
              "addresstype": "suburb",
              "coordinates": "48.8588897,2.3200410",
              "display_name": "Paris, Île-de-France, France métropolitaine",
              "name": "Paris",
              "osm_id": "r7444",
              }
          },
          "status": "success"
        }

    - returning no location

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
            "data": {
                "location": {}
            },
            "status": "success"
        }

    :query string osm_id: OSM id, prefixed by location type
    :query string language: preferred language for Nominatim results

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 500:
        - ``error when getting location from OSM id``

    """
    params = request.args.copy()
    osm_id = params.get("osm_id")
    language = params.get("language")

    location = {}
    if osm_id:
        try:
            # lookup is case-insensitive
            location = nominatim_service.get_location_from_id(
                osm_id.lower(), language
            )
        except Exception:
            message = "error when getting location from OSM id"
            appLog.exception(message)
            return InternalServerErrorResponse(message)

    return {"status": "success", "location": location}



from typing import List


class InvalidEquipmentException(Exception):
    def __init__(
        self, status: str, message: str, equipment_short_id: str
    ) -> None:
        super().__init__(message)
        self.status = status
        self.message = message
        self.equipment_id = equipment_short_id


class InvalidEquipmentsException(Exception):
    pass


class EquipmentForbiddenException(Exception):
    pass


class MiscEquipmentLimitExceededException(Exception):
    def __init__(self, message: str, sport_ids: List[int]) -> None:
        super().__init__(message)
        self.message = message
        self.sport_ids = sport_ids



from datetime import datetime, timezone
from typing import Dict, Union

from flask import Blueprint, current_app, request
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound

from fittrackee import db
from fittrackee.database import PSQL_INTEGER_LIMIT
from fittrackee.oauth2.server import require_auth
from fittrackee.responses import (
    HttpResponse,
    InvalidConfigValueErrorResponse,
    InvalidPayloadErrorResponse,
    handle_error_and_return_response,
)
from fittrackee.users.models import User
from fittrackee.users.roles import UserRole
from fittrackee.users.utils.controls import is_valid_email

from .models import AppConfig
from .utils import update_app_config_from_database, verify_app_config

config_blueprint = Blueprint("config", __name__)

MAX_GLOBAL_MAP_WORKOUTS = 50000  # limitation on browser side


@config_blueprint.route("/config", methods=["GET"])
def get_application_config() -> Union[Dict, HttpResponse]:
    """
    Get Application configuration.

    **Example request**:

    .. sourcecode:: http

      GET /api/config HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "about": null,
          "admin_contact": "admin@example.com",
          "elevation_services":	{
            "open_elevation": false,
            "valhalla": false
          },
          "file_sync_limit_import": 10,
          "file_limit_import": 10,
          "global_map_workouts_limit": 10000,
          "is_email_sending_enabled": true,
          "is_registration_enabled": false,
          "max_single_file_size": 1048576,
          "max_users": 0,
          "max_zip_file_size": 10485760,
          "map_attribution": "&copy; <a href=http://www.openstreetmap.org/copyright>OpenStreetMap</a> contributors",
          "privacy_policy": null,
          "privacy_policy_date": null,
          "stats_workouts_limit": 10000,
          "version": "1.2.1",
          "weather_provider": null
        },
        "status": "success"
      }

    :statuscode 200: ``success``
    :statuscode 500: ``error on getting configuration``
    """

    try:
        config = AppConfig.query.one()
        return {"status": "success", "data": config.serialize()}
    except (MultipleResultsFound, NoResultFound) as e:
        return handle_error_and_return_response(
            e, message="error on getting configuration"
        )


@config_blueprint.route("/config", methods=["PATCH"])
@require_auth(scopes=["application:write"], role=UserRole.ADMIN)
def update_application_config(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    Update Application configuration.

    **Scope**: ``application:write``

    **Minimum role**: Administrator

    **Example request**:

    .. sourcecode:: http

      GET /api/config HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "about": null,
          "admin_contact": "admin@example.com",
          "elevation_services":	{
            "open_elevation": false,
            "valhalla": false
          },
          "file_sync_limit_import": 10,
          "file_limit_import": 10,
          "global_map_workouts_limit": 10000,
          "is_email_sending_enabled": true,
          "is_registration_enabled": false,
          "max_single_file_size": 1048576,
          "max_users": 10,
          "max_zip_file_size": 10485760,
          "map_attribution": "&copy; <a href=http://www.openstreetmap.org/copyright>OpenStreetMap</a> contributors",
          "privacy_policy": null,
          "privacy_policy_date": null,
          "stats_workouts_limit": 10000,
          "version": "1.2.1",
          "weather_provider": null
        },
        "status": "success"
      }

    :<json string about: instance information
    :<json string admin_contact: email to contact the administrator
    :<json integer file_sync_limit_import: max number of files in zip archive,
                   processed synchronously (it must not exceed
                   ``file_limit_import``)
    :<json integer file_limit_import: max number of files in zip archive
    :<json integer global_map_workouts_limit: max number of workouts displayed
                   on global map
    :<json boolean is_registration_enabled: is registration enabled?
    :<json integer max_single_file_size: max size of a single file
    :<json integer max_users: max users allowed to register on instance
    :<json integer max_zip_file_size: max size of a zip archive
    :<json string privacy_policy: instance privacy policy
    :<json integer stats_workouts_limit: max number of workouts for sport
                   statistics

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid payload``
        - ``max size of zip archive must be greater than 0``
        - ``max size of zip archive must be equal or greater than max size of uploaded files``
        - ``max size of uploaded files must be greater than 0``
        - ``max files in a zip archive must be greater than 0``
        - ``max files in a zip archive processed synchronously must be greater than 0``
        - ``max files in a zip archive must be equal or greater than max files in a zip archive processed synchronously``
        - ``max users must be greater than or equal to 0``
        - ``max number of workouts for statistics must be greater than or equal to 0``
        - ``valid email must be provided for admin contact``
        - ``'file_sync_limit_import' must be less than 2147483647``
        - ``'file_limit_import' must be less than 2147483647``
        - ``'max_single_file_size' must be less than 2147483647``
        - ``'max_zip_file_size' must be less than 2147483647``
        - ``'max_users' must be less than 2147483647``
        - ``'stats_workouts_limit' must be less than 2147483647``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
        - ``valid email must be provided for admin contact``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 500: ``error when updating configuration``
    """
    config_data = request.get_json()
    if not config_data:
        return InvalidPayloadErrorResponse()

    ret = verify_app_config(config_data)
    admin_contact = config_data.get("admin_contact")
    if admin_contact and not is_valid_email(admin_contact):
        ret.append("valid email must be provided for admin contact")
    if ret:
        return InvalidPayloadErrorResponse(message=ret)

    try:
        config = AppConfig.query.one()

        for param in [
            "file_sync_limit_import",
            "file_limit_import",
            "max_single_file_size",
            "max_zip_file_size",
            "max_users",
            "stats_workouts_limit",
        ]:
            if param in config_data:
                if (
                    isinstance(config_data[param], int)
                    and config_data[param] > PSQL_INTEGER_LIMIT
                ):
                    return InvalidConfigValueErrorResponse(
                        param, PSQL_INTEGER_LIMIT + 1
                    )
                setattr(config, param, config_data[param])
        if "admin_contact" in config_data:
            config.admin_contact = admin_contact if admin_contact else None
        if "about" in config_data:
            config.about = (
                config_data.get("about") if config_data.get("about") else None
            )
        if "privacy_policy" in config_data:
            privacy_policy = config_data.get("privacy_policy")
            config.privacy_policy = privacy_policy if privacy_policy else None
            config.privacy_policy_date = (
                datetime.now(timezone.utc) if privacy_policy else None
            )
        if "global_map_workouts_limit" in config_data:
            if (
                config_data["global_map_workouts_limit"]
                > MAX_GLOBAL_MAP_WORKOUTS
            ):
                return InvalidPayloadErrorResponse(
                    f"'global_map_workouts_limit' must be less "
                    f"than {MAX_GLOBAL_MAP_WORKOUTS}"
                )
            config.global_map_workouts_limit = config_data[
                "global_map_workouts_limit"
            ]

        if config.max_zip_file_size < config.max_single_file_size:
            return InvalidPayloadErrorResponse(
                "max size of zip archive must be equal or greater than "
                "max size of uploaded files"
            )
        if config.file_limit_import < config.file_sync_limit_import:
            return InvalidPayloadErrorResponse(
                "max files in a zip archive must be equal or greater than "
                "max files in a zip archive processed synchronously"
            )
        db.session.commit()
        update_app_config_from_database(current_app, config)
        return {"status": "success", "data": config.serialize()}

    except Exception as e:
        return handle_error_and_return_response(
            e, message="error when updating configuration"
        )


@config_blueprint.route("/ping", methods=["GET"])
def health_check() -> Union[Dict, HttpResponse]:
    """health check endpoint

    **Example request**:

    .. sourcecode:: http

      GET /api/ping HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "message": "pong!",
        "status": "success"
      }

    :statuscode 200: ``success``
    """
    return {"status": "success", "message": "pong!"}



"""add messages preferences

Revision ID: 141d3978536a
Revises: b30c7818c902
Create Date: 2025-10-01 15:28:59.368811

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '141d3978536a'
down_revision = 'b30c7818c902'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('messages_preferences', postgresql.JSONB(astext_type=sa.Text()), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.drop_column('messages_preferences')

    # ### end Alembic commands ###



import logging
import os
import re
from importlib import import_module, reload
from typing import TYPE_CHECKING, Any, Dict, Tuple

import redis
from dramatiq.brokers.stub import StubBroker
from dramatiq_abort import Abortable, backends
from flask import (
    Flask,
    Response,
    render_template,
    send_file,
    send_from_directory,
)
from flask_babel import Babel
from flask_bcrypt import Bcrypt
from flask_limiter import Limiter
from flask_limiter.errors import RateLimitExceeded
from flask_limiter.util import get_remote_address
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.orm import DeclarativeBase
from werkzeug.middleware.proxy_fix import ProxyFix

from fittrackee.dramatiq_broker import init_dramatiq_broker
from fittrackee.emails.emails import EmailService
from fittrackee.exceptions import EmailConfigException
from fittrackee.request import CustomRequest

VERSION = __version__ = "1.2.1"
DEFAULT_PRIVACY_POLICY_DATA = "2024-12-23 19:00:00"
REDIS_URL = os.getenv("REDIS_URL", "redis://")
API_RATE_LIMITS = os.environ.get("API_RATE_LIMITS")

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(logging.Formatter(fmt="%(message)s"))
log_handlers: list = [stream_handler]

log_file = os.getenv("APP_LOG")
if log_file:
    file_handler = logging.FileHandler(log_file)
    file_handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            datefmt="%Y/%m/%d %H:%M:%S",
        )
    )
    log_handlers.extend([logging.FileHandler(log_file)])

logging.basicConfig(handlers=log_handlers)
appLog = logging.getLogger("fittrackee")


class Base(DeclarativeBase):
    pass


db = SQLAlchemy(
    model_class=Base,
    engine_options={"future": True},
    session_options={"future": True},
)

# workaround with mypy
# see https://github.com/pallets-eco/flask-sqlalchemy/issues/1327
if TYPE_CHECKING:
    from flask_sqlalchemy.model import Model

    BaseModel = Model
else:
    BaseModel = db.Model

babel = Babel()
bcrypt = Bcrypt()
migrate = Migrate()
email_service = EmailService()
redis_client = redis.from_url(REDIS_URL)
redis_available = True
try:
    redis_client.ping()
except redis.exceptions.ConnectionError:
    redis_available = False

limiter = Limiter(
    key_func=get_remote_address,
    default_limits=(
        API_RATE_LIMITS.split(",")  # type: ignore
        if isinstance(API_RATE_LIMITS, str)
        else None
    ),
    default_limits_per_method=True,
    headers_enabled=True,
    storage_uri=REDIS_URL,
    strategy="fixed-window",
)
if not API_RATE_LIMITS:
    limiter.enabled = False
elif not redis_available:
    limiter.enabled = False
    appLog.warning("Redis not available, API rate limits are disabled.")

backend = backends.RedisBackend(client=redis_client)
abortable = Abortable(backend=backend)


class CustomFlask(Flask):
    # add custom Request to handle user-agent parsing
    # (removed in Werkzeug 2.1)
    request_class = CustomRequest


def create_app(init_email: bool = True) -> Flask:
    # instantiate the app
    app = CustomFlask(
        __name__, static_folder="dist/static", template_folder="dist"
    )

    # set config
    with app.app_context():
        app_settings = os.getenv(
            "APP_SETTINGS", "fittrackee.config.ProductionConfig"
        )
        if app_settings == "fittrackee.config.TestingConfig":
            # reload config on tests
            config = import_module("fittrackee.config")
            reload(config)
        app.config.from_object(app_settings)

    # set up extensions
    babel.init_app(app)
    db.init_app(app)
    bcrypt.init_app(app)
    migrate.init_app(app, db)
    limiter.init_app(app)

    init_dramatiq_broker(app, abortable, REDIS_URL)

    # set oauth2
    from fittrackee.oauth2.config import config_oauth

    config_oauth(app)

    # check if dramatiq broker is available
    app.config["TASKS_PROCESSING_AVAILABLE"] = (
        issubclass(app.config["DRAMATIQ_BROKER"], StubBroker)  # tests
        or redis_available  # dev/prod
    )

    # set up email if 'EMAIL_URL' is initialized and redis available
    if init_email:
        if app.config["EMAIL_URL"]:
            if not app.config["TASKS_PROCESSING_AVAILABLE"]:
                raise EmailConfigException(
                    "EMAIL_URL is set, but Redis is not available. Please "
                    "check application configuration or Redis status."
                )
            email_service.init_email(app)
            app.config["CAN_SEND_EMAILS"] = True
        else:
            appLog.warning(
                "EMAIL_URL is not provided, email sending is deactivated."
            )

    # get configuration from database
    from .application.utils import (
        get_or_init_config,
        update_app_config_from_database,
    )

    with app.app_context():
        # Note: check if "app_config" table exist to avoid errors when
        # dropping tables on dev environments
        try:
            with db.engine.connect() as conn:
                if db.engine.dialect.has_table(conn, "app_config"):
                    db_app_config = get_or_init_config()
                    update_app_config_from_database(app, db_app_config)
        except ProgrammingError as e:
            # avoid error on AppConfig migration
            if re.match(
                r"psycopg2.errors.UndefinedColumn(.*)app_config.", str(e)
            ):
                pass

    from .workouts.tasks import upload_workouts_archive  # noqa

    from .application.app_config import config_blueprint
    from .comments.comments import comments_blueprint
    from .equipments.equipment_types import equipment_types_blueprint
    from .equipments.equipments import equipments_blueprint
    from .feeds.routes import feeds_blueprint
    from .geocode.routes import geocode_blueprint
    from .oauth2.routes import oauth2_blueprint
    from .reports.reports import reports_blueprint
    from .users.auth import auth_blueprint
    from .users.follow_requests import follow_requests_blueprint
    from .users.notifications import notifications_blueprint
    from .users.queued_tasks import queued_tasks_blueprint
    from .users.users import users_blueprint
    from .workouts.records import records_blueprint
    from .workouts.sports import sports_blueprint
    from .workouts.stats import stats_blueprint
    from .workouts.timeline import timeline_blueprint
    from .workouts.workouts import workouts_blueprint

    app.register_blueprint(auth_blueprint, url_prefix="/api")
    app.register_blueprint(equipment_types_blueprint, url_prefix="/api")
    app.register_blueprint(equipments_blueprint, url_prefix="/api")
    app.register_blueprint(oauth2_blueprint, url_prefix="/api")
    app.register_blueprint(comments_blueprint, url_prefix="/api")
    app.register_blueprint(config_blueprint, url_prefix="/api")
    app.register_blueprint(records_blueprint, url_prefix="/api")
    app.register_blueprint(sports_blueprint, url_prefix="/api")
    app.register_blueprint(stats_blueprint, url_prefix="/api")
    app.register_blueprint(queued_tasks_blueprint, url_prefix="/api")
    app.register_blueprint(users_blueprint, url_prefix="/api")
    app.register_blueprint(workouts_blueprint, url_prefix="/api")
    app.register_blueprint(follow_requests_blueprint, url_prefix="/api")
    app.register_blueprint(timeline_blueprint, url_prefix="/api")
    app.register_blueprint(notifications_blueprint, url_prefix="/api")
    app.register_blueprint(reports_blueprint, url_prefix="/api")
    app.register_blueprint(feeds_blueprint, url_prefix="")
    app.register_blueprint(geocode_blueprint, url_prefix="/api")

    if app.debug:
        logging.getLogger("sqlalchemy").setLevel(logging.WARNING)
        logging.getLogger("sqlalchemy").handlers = logging.getLogger(
            "werkzeug"
        ).handlers
        logging.getLogger("sqlalchemy.orm").setLevel(logging.WARNING)
        logging.getLogger("flake8").propagate = False
        appLog.setLevel(logging.DEBUG)

        # Enable CORS
        @app.after_request  # type: ignore
        def after_request(response: Response) -> Response:
            response.headers.add("Access-Control-Allow-Origin", "*")
            response.headers.add(
                "Access-Control-Allow-Headers", "Content-Type,Authorization"
            )
            response.headers.add(
                "Access-Control-Allow-Methods",
                "GET,PUT,POST,DELETE,PATCH,OPTIONS",
            )
            return response

    @app.errorhandler(429)
    def rate_limit_handler(error: RateLimitExceeded) -> Tuple[Dict, int]:
        return {
            "status": "error",
            "message": f"rate limit exceeded ({error.description})",
        }, 429

    @app.route("/favicon.ico")
    @limiter.exempt
    def favicon() -> Any:
        return send_file(
            os.path.join(app.root_path, "dist/favicon.ico")  # type: ignore
        )

    @app.route("/", defaults={"path": ""})
    @app.route("/<path:path>")
    @limiter.exempt
    def catch_all(path: str) -> Any:
        # workaround to serve images (not in static directory)
        if path.startswith("img/"):
            return send_from_directory(
                directory=os.path.join(
                    app.root_path,  # type: ignore
                    "dist",
                ),
                path=path,
            )
        else:
            return render_template("index.html")

    # to get headers, especially 'X-Forwarded-Proto' for scheme needed by
    # Authlib, when the application is running behind a proxy server
    app.wsgi_app = ProxyFix(app.wsgi_app)  # type: ignore

    return app



from threading import local
from typing import TYPE_CHECKING, Any, Optional

import dramatiq
from dramatiq import Middleware
from dramatiq.brokers.redis import RedisBroker
from dramatiq.middleware import Retries, ShutdownNotifications, TimeLimit

if TYPE_CHECKING:
    from dramatiq_abort import Abortable
    from flask import Flask


class AppContextMiddleware(Middleware):
    # From https://github.com/Bogdanp/flask_dramatiq_example

    state = local()

    def __init__(self, app: "Flask") -> None:
        self.app = app

    def before_process_message(
        self,
        broker: "dramatiq.broker.Broker",
        message: "dramatiq.broker.MessageProxy",
    ) -> None:
        context = self.app.app_context()
        context.push()

        self.state.context = context

    def after_process_message(
        self,
        broker: "dramatiq.broker.Broker",
        message: "dramatiq.broker.MessageProxy",
        *,
        result: Any = None,
        exception: Optional[BaseException] = None,
    ) -> None:
        try:
            context = self.state.context
            context.pop(exception)
            del self.state.context
        except AttributeError:
            pass

    after_skip_message = after_process_message


def init_dramatiq_broker(
    app: "Flask", abortable: "Abortable", redis_url: str
) -> None:
    broker_cls = app.config["DRAMATIQ_BROKER"]
    kw = {}
    if issubclass(broker_cls, RedisBroker):
        kw["url"] = redis_url
    middlewares = [
        AppContextMiddleware(app),
        TimeLimit(),
        ShutdownNotifications(),
        Retries(),
        abortable,
    ]
    broker = broker_cls(middleware=middlewares, **kw)

    # At startup, actors are already registered with the default global broker
    # (no lazy loading).
    # This allows to update actors in factory function and use broker from
    # application config
    default_broker = dramatiq.get_broker()
    for existing_actor_name in default_broker.get_declared_actors():
        actor = default_broker.get_actor(existing_actor_name)
        actor.broker = broker
        broker.declare_actor(actor)

    dramatiq.set_broker(broker)



WORKOUT_ALLOWED_EXTENSIONS = {"gpx", "fit", "kml", "kmz", "tcx"}
WORKOUT_DATE_FORMAT = "%Y-%m-%d %H:%M"

# sports with cadence displayed in revolutions per minute
RPM_CADENCE_SPORTS = [
    "Cycling (Sport)",
    "Cycling (Trekking)",
    "Cycling (Transport)",
    "Cycling (Virtual)",
    "Halfbike",
    "Mountain Biking",
    "Mountain Biking (Electric)",
    "Open Water Swimming",
]
# sports with cadence displayed in steps per minute
SPM_CADENCE_SPORTS = [
    "Hiking",
    "Mountaineering",
    "Snowshoes",
    "Running",
    "Trail",
    "Walking",
]
CADENCE_SPORTS = [*RPM_CADENCE_SPORTS, *SPM_CADENCE_SPORTS]
POWER_SPORTS = [
    "Cycling (Sport)",
    "Cycling (Trekking)",
    "Cycling (Transport)",
    "Cycling (Virtual)",
    "Halfbike",
    "Mountain Biking",
    "Mountain Biking (Electric)",
]
PACE_SPORTS = [
    "Hiking",
    "Running",
    "Trail",
    "Walking",
]
# elevation data (ascent, descent, min and max alt), if present, are not
# stored and/or displayed
SPORTS_WITHOUT_ELEVATION_DATA = [
    "Ice Skating",
    # racket sports
    "Padel (Outdoor)",
    "Tennis (Outdoor)",
    # flatwater sports
    "Canoeing",
    "Kayaking",
    "Open Water Swimming",
    "Rowing",
    "Standup Paddleboarding",
]

WORKOUT_FILE_MIMETYPES = {
    "fit": "application/vnd.ant.fit",
    "gpx": "application/gpx+xml",
    "kml": "application/vnd.google-earth.kml+xml",
    "kmz": "application/vnd.google-earth.kmz",
    "tcx": "application/vnd.garmin.tcx+xml",
}

NSMAP = {
    "gpxtpx": "http://www.garmin.com/xmlschemas/TrackPointExtension/v1",
}
TRACK_EXTENSION_NSMAP = {
    "gpxtrkx": "http://www.garmin.com/xmlschemas/TrackStatsExtension/v1"
}

WGS84_CRS = 4326  # World Geodetic System 1984, in degrees



"""update User and AppConfig tables

Revision ID: 5e3a3a31c432
Revises: e30007d681cb
Create Date: 2022-02-23 11:05:24.223304

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '5e3a3a31c432'
down_revision = 'e30007d681cb'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        'users', 'username', existing_type=sa.String(length=20),
        type_=sa.String(length=255), existing_nullable=False
    )
    op.alter_column(
        'users', 'email', existing_type=sa.String(length=120),
        type_=sa.String(length=255), existing_nullable=False
    )
    op.add_column(
        'users',
        sa.Column('is_active', sa.Boolean(), default=False, nullable=True))
    op.execute("UPDATE users SET is_active = true")
    op.alter_column('users', 'is_active', nullable=False)
    op.add_column(
        'users',
        sa.Column('email_to_confirm', sa.String(length=255), nullable=True))
    op.add_column(
        'users',
        sa.Column('confirmation_token', sa.String(length=255), nullable=True))

    op.add_column(
        'app_config',
        sa.Column('admin_contact', sa.String(length=255), nullable=True)
    )


def downgrade():
    op.drop_column('app_config', 'admin_contact')

    op.drop_column('users', 'confirmation_token')
    op.drop_column('users', 'email_to_confirm')
    op.drop_column('users', 'is_active')
    op.alter_column(
        'users', 'email', existing_type=sa.String(length=255),
        type_=sa.String(length=120), existing_nullable=False
    )
    op.alter_column(
        'users', 'username', existing_type=sa.String(length=255),
        type_=sa.String(length=20), existing_nullable=False
    )



import os
import secrets
import zipfile
from abc import abstractmethod
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from io import BytesIO
from typing import IO, TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from flask import current_app

from fittrackee import appLog, db
from fittrackee.equipments.exceptions import InvalidEquipmentsException
from fittrackee.equipments.models import Equipment
from fittrackee.files import get_absolute_file_path
from fittrackee.users.models import Notification, User, UserTask
from fittrackee.visibility_levels import get_calculated_visibility
from fittrackee.workouts.models import (
    DESCRIPTION_MAX_CHARACTERS,
    NOTES_MAX_CHARACTERS,
)

from ..exceptions import (
    WorkoutExceedingValueException,
    WorkoutException,
    WorkoutFileException,
    WorkoutNoFileException,
)
from .base_workout_service import (
    BaseWorkoutService,
)
from .mixins import WorkoutFileMixin
from .workout_from_file.services import WORKOUT_FROM_FILE_SERVICES

if TYPE_CHECKING:
    from werkzeug.datastructures import FileStorage

    from fittrackee.visibility_levels import VisibilityLevel
    from fittrackee.workouts.models import Workout


@dataclass
class WorkoutsData:
    sport_id: int
    analysis_visibility: Optional["VisibilityLevel"] = None
    description: Optional[str] = None
    equipment_ids: Optional[List[str]] = None
    map_visibility: Optional["VisibilityLevel"] = None
    notes: Optional[str] = None
    title: Optional[str] = None
    workout_visibility: Optional["VisibilityLevel"] = None


class AbstractWorkoutsCreationService(BaseWorkoutService, WorkoutFileMixin):
    def __init__(
        self,
        auth_user: "User",
        workouts_data: Dict,
        file: Optional["FileStorage"] = None,
    ):
        super().__init__(
            auth_user,
            workouts_data.get("sport_id"),
            workouts_data.get("equipment_ids"),
        )
        self.file: Optional["FileStorage"] = file
        self.workouts_data = WorkoutsData(**workouts_data)

    @abstractmethod
    def process(self) -> Tuple[List["Workout"], Dict]:
        pass

    def _get_user_path(
        self,
    ) -> str:
        user_path = os.path.join("workouts", str(self.auth_user.id))
        absolute_user_path = get_absolute_file_path(user_path)
        if not os.path.exists(absolute_user_path):
            os.makedirs(absolute_user_path)
        return user_path

    def get_file_path(
        self,
        workout_date: str,
        extension: str,
    ) -> str:
        suffix = secrets.token_urlsafe(8)
        new_filename = (
            f"{workout_date}_{self.workouts_data.sport_id}_{suffix}{extension}"
        )
        dir_path = self._get_user_path()
        file_path = os.path.join(dir_path, new_filename.split("/")[-1])
        return file_path

    def _get_workout_title(
        self, service_workout_name: Optional[str], workout_date: datetime
    ) -> Optional[str]:
        title = ""
        if self.workouts_data.title:
            title = self.workouts_data.title
        elif service_workout_name:
            title = service_workout_name
        return self._get_title(workout_date, title)

    def _get_workout_description(
        self, service_workout_description: Optional[str]
    ) -> Optional[str]:
        if self.workouts_data.description:
            return self.workouts_data.description[:DESCRIPTION_MAX_CHARACTERS]

        if service_workout_description:
            return service_workout_description[:DESCRIPTION_MAX_CHARACTERS]

        return None

    def _get_workout_notes(self) -> Optional[str]:
        if self.workouts_data.notes is None:
            return None
        return self.workouts_data.notes[:NOTES_MAX_CHARACTERS]

    def _store_file(
        self,
        new_workout: "Workout",
        workout_file: Optional["IO[bytes]"],
        *,
        extension: str,
    ) -> str:
        if not workout_file and not self.file:
            return ""

        relative_path = self.get_file_path(
            workout_date=new_workout.workout_date.strftime(
                "%Y-%m-%d_%H-%M-%S"
            ),
            extension=extension,
        )

        absolute_workout_filepath = get_absolute_file_path(relative_path)
        try:
            if workout_file:  # when file from archive
                with open(absolute_workout_filepath, "wb") as f:
                    workout_file.seek(0)
                    f.write(workout_file.read())
            elif self.file:
                self.file.stream.seek(0)
                self.file.save(absolute_workout_filepath)
            else:
                return ""
        except Exception as e:
            error = "error when storing workout file"
            appLog.exception(error)
            raise WorkoutException("error", error) from e

        new_workout.original_file = relative_path
        return absolute_workout_filepath

    def _get_archive_content(self) -> Union[BytesIO, IO[bytes]]:
        if not self.file:
            raise WorkoutNoFileException()

        # handle Python < 3.11, see:
        # - https://github.com/python/cpython/issues/70363
        # - https://docs.python.org/3.11/whatsnew/3.11.html#tempfile
        if not hasattr(self.file.stream, "seekable"):
            return BytesIO(self.file.getvalue())
        return self.file.stream

    def create_workout_from_file(
        self,
        extension: str,
        equipments: Union[List["Equipment"], None],
        workout_file: Optional["IO[bytes]"] = None,
        get_weather: bool = True,
    ) -> "Workout":
        """
        Return map absolute file path in order to delete file on error
        """
        if workout_file is None and self.file is None:
            raise WorkoutNoFileException()

        if extension == "kmz" and workout_file is None:
            workout_file = self._get_archive_content()

        workout_service = WORKOUT_FROM_FILE_SERVICES[extension](
            auth_user=self.auth_user,
            workout_file=(
                self.file.stream  # type: ignore
                if workout_file is None
                else workout_file
            ),
            sport=self.sport,
            stopped_speed_threshold=self.stopped_speed_threshold,
            get_weather=get_weather,
            workout=None,
        )

        # extract and calculate data from provided file
        try:
            new_workout = workout_service.process_workout()
        except (WorkoutExceedingValueException, WorkoutFileException) as e:
            appLog.exception(f"workout exception: {e!s}")
            db.session.rollback()
            raise e
        except Exception as e:
            appLog.exception(f"exception: {e!s}")
            db.session.rollback()
            if (
                "duplicate key value violates unique constraint "
                '"workout_id_start_date_unique"' in str(e)
            ):
                raise WorkoutFileException(
                    "error", "some segments have same start date"
                ) from e
            raise WorkoutException(
                "error", "error when processing workout"
            ) from e

        # store title, description and notes
        new_workout.title = self._get_workout_title(
            workout_service.workout_name, new_workout.workout_date
        )
        new_workout.description = self._get_workout_description(
            workout_service.workout_description
        )
        new_workout.notes = self._get_workout_notes()

        # add equipments if ids provided
        if equipments is not None:
            new_workout.equipments = equipments

        # update visibility
        new_workout.workout_visibility = (
            self.workouts_data.workout_visibility
            if self.workouts_data.workout_visibility
            else self.auth_user.workouts_visibility
        )
        new_workout.analysis_visibility = get_calculated_visibility(
            visibility=(
                self.workouts_data.analysis_visibility
                if self.workouts_data.analysis_visibility
                else self.auth_user.analysis_visibility
            ),
            parent_visibility=new_workout.workout_visibility,
        )
        new_workout.map_visibility = get_calculated_visibility(
            visibility=(
                self.workouts_data.map_visibility
                if self.workouts_data.map_visibility
                else self.auth_user.map_visibility
            ),
            parent_visibility=new_workout.analysis_visibility,
        )

        # store original workout file
        absolute_workout_filepath = self._store_file(
            new_workout,
            workout_file,
            extension=f".{extension}",
        )

        # generate and store map image
        map_filepath = self.get_file_path(
            workout_date=new_workout.workout_date.strftime(
                "%Y-%m-%d_%H-%M-%S"
            ),
            extension=".png",
        )
        new_workout.map = map_filepath
        absolute_map_filepath = get_absolute_file_path(map_filepath)
        try:
            workout_service.generate_map_image(
                map_filepath=absolute_map_filepath,
                coordinates=workout_service.coordinates,
            )
            new_workout.map_id = workout_service.get_map_hash(map_filepath)
        except Exception as e:
            if os.path.exists(absolute_map_filepath):
                os.remove(absolute_map_filepath)
            if os.path.exists(absolute_workout_filepath):
                os.remove(absolute_workout_filepath)
            raise WorkoutException(
                "error", "error when generating map image"
            ) from e
        db.session.commit()
        return new_workout

    def process_archive_content(
        self,
        archive_content: Union[BytesIO, IO[bytes]],
        files_to_process: List[str],
        equipments: Union[List["Equipment"], None],
        upload_task: Optional["UserTask"] = None,
        get_weather: bool = True,
    ) -> Tuple[List["Workout"], Dict]:
        if not files_to_process:
            raise WorkoutFileException(
                "error", "No files from archive to process"
            )
        appLog.debug(" > starting archive processing...")

        new_workouts = []
        errored_workouts = {}
        total_files = len(files_to_process)
        with zipfile.ZipFile(archive_content, "r") as zip_ref:
            for index, file in enumerate(files_to_process, start=1):
                appLog.debug(f"  - file {index}/{total_files}")
                try:
                    extension = self._get_extension(file)
                    file_content = zip_ref.open(file)
                    new_workout = self.create_workout_from_file(
                        extension, equipments, file_content, get_weather
                    )
                except Exception as e:
                    db.session.rollback()
                    error = e.args[0]
                    errored_workouts[file] = error
                    appLog.debug(f"    > error occurred: {error}")
                    if upload_task:
                        upload_task.progress = int(100 * index / total_files)
                        db.session.commit()
                    continue

                new_workouts.append(new_workout)
                appLog.debug("    > upload done")
                if upload_task:
                    upload_task.data = {
                        **upload_task.data,
                        "new_workouts_count": len(new_workouts),
                    }
                    upload_task.progress = int(100 * index / total_files)
                    db.session.commit()

        return new_workouts, errored_workouts


class WorkoutsFromFileCreationService(AbstractWorkoutsCreationService):
    def __init__(
        self,
        auth_user: "User",
        workouts_data: Dict,
        file: Optional["FileStorage"] = None,
    ):
        super().__init__(auth_user, workouts_data, file)

    def get_files_from_archive(self) -> List[str]:
        if not self.file:
            raise WorkoutNoFileException()
        files_to_process = []
        max_file_size = current_app.config["max_single_file_size"]
        try:
            with zipfile.ZipFile(self.file.stream, "r") as zip_ref:
                for zip_info in zip_ref.infolist():
                    if self._is_workout_file(zip_info.filename):
                        if zip_info.file_size > max_file_size:
                            raise WorkoutFileException(
                                "fail",
                                (
                                    "at least one file in zip archive "
                                    "exceeds size limit, "
                                    "please check the archive"
                                ),
                            )
                        files_to_process.append(zip_info.filename)

                        if (
                            len(files_to_process)
                            > current_app.config["file_limit_import"]
                        ):
                            raise WorkoutFileException(
                                "fail",
                                (
                                    "the number of files in the archive "
                                    "exceeds the limit"
                                ),
                            )

        except zipfile.BadZipFile as e:
            raise WorkoutFileException("error", "invalid zip file") from e

        if not files_to_process:
            raise WorkoutFileException(
                "fail", "archive does not contain valid workout files"
            )

        return files_to_process

    def add_workouts_upload_task(
        self,
        files_to_process: List[str],
        equipments: Union[List["Equipment"], None],
    ) -> str:
        from fittrackee.workouts.tasks import upload_workouts_archive

        if self.file is None:
            raise WorkoutNoFileException()
        if not files_to_process:
            raise WorkoutFileException(
                "error", "No files from archive to process"
            )
        if UserTask.query.filter(
            UserTask.user_id == self.auth_user.id,
            UserTask.task_type == "workouts_archive_upload",
            UserTask.aborted == False,  # noqa
            UserTask.errored == False,  # noqa
            UserTask.progress != 100,
        ).first():
            raise WorkoutException("invalid", "ongoing upload task exists")

        relative_path = os.path.join(
            self._get_user_path(),
            f"archive_{secrets.token_urlsafe(20)}.zip",
        )
        path = get_absolute_file_path(relative_path)
        self.file.stream.seek(0)
        self.file.save(path)

        upload_task = UserTask(
            user_id=self.auth_user.id,
            task_type="workouts_archive_upload",
            data={
                "workouts_data": asdict(self.workouts_data),
                "files_to_process": files_to_process,
                "new_workouts_count": 0,
                "equipment_ids": (
                    None
                    if equipments is None
                    else [equipment.id for equipment in equipments]
                ),
                "original_file_name": self.file.filename,
            },
            file_path=relative_path,
        )
        upload_task.errors = {
            "archive": None,
            "files": {},
        }
        upload_task.file_size = os.path.getsize(path)
        db.session.add(upload_task)
        db.session.commit()

        if current_app.config["TASKS_PROCESSING_AVAILABLE"]:
            message = upload_workouts_archive.send(task_id=upload_task.id)
            upload_task.message_id = message.message_id
            db.session.commit()

        return upload_task.short_id

    def process_zip_archive(
        self, equipments: Union[List["Equipment"], None]
    ) -> Tuple[List["Workout"], Dict]:
        if not self.file:
            raise WorkoutNoFileException()
        files_to_process = self.get_files_from_archive()
        if (
            len(files_to_process)
            > current_app.config["file_sync_limit_import"]
        ):
            task_short_id = self.add_workouts_upload_task(
                files_to_process, equipments
            )
            return [], {"task_short_id": task_short_id}

        new_workouts, errored_workouts = self.process_archive_content(
            archive_content=self._get_archive_content(),
            files_to_process=files_to_process,
            equipments=equipments,
        )
        return new_workouts, {
            "errored_workouts": errored_workouts,
            "task_short_id": None,
        }

    def process(self) -> Tuple[List["Workout"], Dict]:
        if self.file is None:
            raise WorkoutNoFileException()
        if self.file.filename is None:
            raise WorkoutFileException("error", "workout file has no filename")

        extension = self._get_extension(self.file.filename)
        if extension not in current_app.config["WORKOUT_ALLOWED_EXTENSIONS"]:
            raise WorkoutFileException(
                "error", "workout file invalid extension"
            )
        equipments = self.get_equipments()

        try:
            if extension == "zip":
                new_workouts, processing_output = self.process_zip_archive(
                    equipments
                )
                return new_workouts, processing_output
            else:
                new_workout = self.create_workout_from_file(
                    extension, equipments
                )
                return [new_workout], {
                    "task_short_id": None,
                    "errored_workouts": [],
                }
        except InvalidEquipmentsException as e:
            raise WorkoutException("error", str(e)) from e


class WorkoutsFromArchiveCreationAsyncService(AbstractWorkoutsCreationService):
    def __init__(
        self,
        task_id: int,
    ) -> None:
        upload_task = UserTask.query.filter_by(id=task_id).first()
        if (
            not upload_task
            or upload_task.task_type != "workouts_archive_upload"
        ):
            raise WorkoutException("error", "no import task found")
        auth_user = User.query.filter_by(id=upload_task.user_id).one()
        import_data = upload_task.data
        super().__init__(auth_user, import_data["workouts_data"], None)
        self.file_path = get_absolute_file_path(upload_task.file_path)
        self.files_to_process = import_data["files_to_process"]
        self.equipment_ids = import_data["equipment_ids"]
        self.upload_task = upload_task

    def process(self) -> Tuple[List["Workout"], Dict]:
        if self.equipment_ids is None:
            equipments = None
        elif not self.equipment_ids:
            equipments = []
        else:
            equipments = Equipment.query.filter(
                Equipment.id.in_(self.equipment_ids)
            ).all()

        errored_workouts: Dict = {}
        new_workouts: List["Workout"] = []
        try:
            with open(self.file_path, "rb") as zip_file:
                new_workouts, errored_workouts = self.process_archive_content(
                    archive_content=zip_file,
                    files_to_process=self.files_to_process,
                    equipments=equipments,
                    upload_task=self.upload_task,
                    get_weather=False,
                )
        except FileNotFoundError:
            self.upload_task.errored = True
            self.upload_task.errors = {
                **self.upload_task.errors,
                "archive": "archive file does not exist",
            }
        except Exception:
            self.upload_task.errored = True
            self.upload_task.errors = {
                **self.upload_task.errors,
                "archive": "error during archive processing",
            }

        if errored_workouts:
            self.upload_task.errored = True
            self.upload_task.errors = {
                **self.upload_task.errors,
                "files": errored_workouts,
            }
        db.session.commit()

        if os.path.exists(self.file_path):
            os.remove(self.file_path)

        notification = Notification(
            from_user_id=self.upload_task.user_id,
            to_user_id=self.upload_task.user_id,
            created_at=datetime.now(tz=timezone.utc),
            event_object_id=self.upload_task.id,
            event_type=self.upload_task.task_type,
        )
        db.session.add(notification)
        db.session.commit()

        return new_workouts, {
            "errored_workouts": errored_workouts,
            "task_short_id": self.upload_task.short_id,
        }



import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List

import numpy as np

from fittrackee import appLog

from .exceptions import ElevationServiceException

if TYPE_CHECKING:
    from gpxpy.gpx import GPXTrackPoint

WINDOW_LEN = 51


class BaseElevationService(ABC):
    config_key: str = ""
    url_pattern: str = ""
    log_label: str = ""

    def __init__(self) -> None:
        self.url = self._get_api_url()

    @property
    def is_enabled(self) -> bool:
        return self.url != ""

    def _get_api_url(self) -> str:
        base_url = os.environ.get(self.config_key)
        if not base_url:
            return ""
        return self.url_pattern.format(base_url=base_url)

    @staticmethod
    def smooth_elevations(points: List[int]) -> List[int]:
        """
        smooth elevations using 'flat' window

        based on SciPy Cookbook:
        https://scipy-cookbook.readthedocs.io/items/SignalSmooth.html
        """
        if len(points) < 3:
            return [int(p) for p in points]

        points_array = np.array(points)
        window_len = len(points) if len(points) < WINDOW_LEN else WINDOW_LEN

        s = np.r_[
            points_array[window_len - 1 : 0 : -1],
            points_array,
            points_array[-2 : -window_len - 1 : -1],
        ]
        w = np.ones(window_len, "d")
        y = np.convolve(w / w.sum(), s, mode="valid")
        start = window_len // 2 + 1
        end = start + len(points_array)
        smooth_array = y[start:end]

        return [int(p) for p in smooth_array]

    @abstractmethod
    def _get_elevations_for_api(
        self, points: List["GPXTrackPoint"], smooth: bool = False
    ) -> List[int]:
        pass

    def get_elevations(
        self, points: List["GPXTrackPoint"], smooth: bool = False
    ) -> List[int]:
        appLog.debug(
            "{log_label}: getting elevations".format(log_label=self.log_label)
        )

        results = self._get_elevations_for_api(points)

        # Should not happen
        if len(results) != len(points):
            error = (
                f"{self.log_label}: mismatch between number of points in "
                "results"
            )
            appLog.error(error)
            raise ElevationServiceException(error)

        if smooth:
            return self.smooth_elevations(results)
        return results



"""add skating sports

Revision ID: 84394acbebcf
Revises: 076d33eab04f
Create Date: 2026-03-04 16:21:27.343612

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '84394acbebcf'
down_revision = '076d33eab04f'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Ice Skating', True, 1.0), ('Inline Skating', True, 1.0)
        """
    )
    op.execute(
        """
        INSERT INTO equipment_types (label, is_active)
        VALUES ('Skates', True);
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM equipment_types
        WHERE label IN ('Skates');
        """
    )
    op.execute(
        """
        DELETE FROM sports
        WHERE label IN ('Ice Skating', 'Inline Skating');
        """
    )




"""add padel (outdoor)

Revision ID: 67ae8866b18f
Revises: 0a2e2369b972
Create Date: 2025-09-24 13:13:35.236703

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '67ae8866b18f'
down_revision = '0a2e2369b972'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Padel (Outdoor)', True, 0.1);
        """
    )


def downgrade():
    op.execute(
        """
        DELETE
        FROM sports
        WHERE label = 'Padel (Outdoor)';
        """
    )



"""add swimrun

Revision ID: c9c6a5a4dd6d
Revises: 032ba0846eea
Create Date: 2024-08-24 08:37:37.751827

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c9c6a5a4dd6d'
down_revision = '032ba0846eea'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Swimrun', True, 0.1)
        """
    )

def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Swimrun';
        """
    )



from fittrackee.utils import clean


def clean_tokens(days: int) -> int:
    sql = """
        DELETE FROM oauth2_token
        WHERE oauth2_token.issued_at + oauth2_token.expires_in < :limit;
    """
    return clean(sql, days)



from .workout_creation_service import WorkoutCreationService
from .workout_from_file import (
    WorkoutFitService,
    WorkoutGpxService,
    WorkoutKmlService,
    WorkoutKmzService,
    WorkoutTcxService,
)
from .workout_update_service import WorkoutUpdateService
from .workouts_from_file_creation_service import (
    WorkoutsFromArchiveCreationAsyncService,
    WorkoutsFromFileCreationService,
)

__all__ = [
    "WorkoutCreationService",
    "WorkoutFitService",
    "WorkoutGpxService",
    "WorkoutKmlService",
    "WorkoutKmzService",
    "WorkoutTcxService",
    "WorkoutUpdateService",
    "WorkoutsFromArchiveCreationAsyncService",
    "WorkoutsFromFileCreationService",
]



"""add windsurfing sport and board equipment

Revision ID: 514c727b7feb
Revises: 61ecbc76f410
Create Date: 2025-05-10 10:45:00.915019

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '514c727b7feb'
down_revision = '61ecbc76f410'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Windsurfing', True, 1);
        """
    )
    op.execute(
        """
        INSERT INTO equipment_types (label, is_active)
        VALUES ('Board', True);
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM equipment_types
        WHERE label = 'Board';
        """
    )
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Windsurfing';
        """
    )



from typing import TYPE_CHECKING

from flask import current_app

from ..exceptions import WorkoutExceedingValueException
from ..models import PSQL_INTEGER_LIMIT, WORKOUT_VALUES_LIMIT
from ..utils.gpx import get_file_extension

if TYPE_CHECKING:
    from fittrackee.workouts.models import Workout


class CheckWorkoutMixin:
    @staticmethod
    def _check_workout(workout: "Workout") -> None:
        if workout.distance and workout.distance > 999.9:
            raise WorkoutExceedingValueException(
                "'distance' exceeds max value (999.9)"
            )
        if workout.duration.total_seconds() > PSQL_INTEGER_LIMIT:
            raise WorkoutExceedingValueException(
                f"'duration' exceeds max value ({PSQL_INTEGER_LIMIT})"
            )

        for key in ["ascent", "descent", "max_speed", "calories"]:
            workout_value = getattr(workout, key)
            max_value = WORKOUT_VALUES_LIMIT[key]
            if workout_value and workout_value > max_value:
                raise WorkoutExceedingValueException(
                    f"'{key}' exceeds max value ({max_value})"
                )


class WorkoutFileMixin:
    @staticmethod
    def _get_extension(filename: str) -> str:
        return get_file_extension(filename)

    @staticmethod
    def _is_valid_workout_file_extension(extension: str) -> bool:
        return (
            extension in current_app.config["WORKOUT_ALLOWED_EXTENSIONS"]
            and extension != "zip"
        )

    def _is_workout_file(self, filename: str) -> bool:
        extension = self._get_extension(filename)
        return self._is_valid_workout_file_extension(extension)



import logging
import sys
from datetime import datetime
from logging import Logger
from typing import Optional

import click

from fittrackee.cli.app import app
from fittrackee.users.models import User
from fittrackee.workouts.constants import WORKOUT_ALLOWED_EXTENSIONS
from fittrackee.workouts.models import Sport
from fittrackee.workouts.services.workouts_from_file_refresh_service import (
    WorkoutsFromFileRefreshService,
)
from fittrackee.workouts.services.workouts_without_file_refresh_service import (  # noqa
    WorkoutsWithoutFileRefreshService,
)
from fittrackee.workouts.tasks import (
    process_workouts_archive_upload,
    process_workouts_archives_uploads,
)
from fittrackee.workouts.utils.workouts import get_workout_datetime

WORKOUT_VALID_EXTENSIONS = ", ".join(WORKOUT_ALLOWED_EXTENSIONS)
VALID_ON_ERROR_CHOICES = ["remove-references", "delete-workout"]

logger = logging.getLogger("fittrackee_workouts_cli")
logger.setLevel(logging.INFO)


def validate_order(
    ctx: click.core.Context, param: click.core.Option, value: Optional[str]
) -> Optional[str]:
    if value and value not in ["asc", "desc"]:
        raise click.BadParameter("order must be 'asc' or 'desc'")
    return value


def validate_extension(
    ctx: click.core.Context, param: click.core.Option, value: Optional[str]
) -> Optional[str]:
    if value and value not in WORKOUT_ALLOWED_EXTENSIONS:
        raise click.BadParameter(
            f"valid extensions are: {WORKOUT_VALID_EXTENSIONS}"
        )
    return value


def validate_sport_id(
    ctx: click.core.Context, param: click.core.Option, value: Optional[int]
) -> Optional[int]:
    with app.app_context():
        if value is not None and not Sport.query.filter_by(id=value).first():
            raise click.BadParameter(f"invalid sport id '{value}'")
    return value


def validate_user(
    ctx: click.core.Context, param: click.core.Option, value: Optional[str]
) -> Optional[str]:
    with app.app_context():
        if (
            value is not None
            and not User.query.filter_by(username=value).first()
        ):
            raise click.BadParameter(f"user '{value}' does not exist")
    return value


def validate_number(
    ctx: click.core.Context, param: click.core.Option, value: Optional[int]
) -> Optional[int]:
    if value is not None and value < 1:
        raise click.BadParameter("value must be greater than 0")

    return value


def validate_date(
    ctx: click.core.Context, param: click.core.Option, value: Optional[str]
) -> Optional["datetime"]:
    if value is not None:
        try:
            date_value, _ = get_workout_datetime(
                workout_date=value,
                user_timezone=None,
                date_str_format="%Y-%m-%d",
            )
            return date_value
        except Exception as e:
            raise click.BadParameter(
                f"'{value}' does not match format '%Y-%m-%d'"
            ) from e
    return value


def validate_on_file_error(
    ctx: click.core.Context, param: click.core.Option, value: Optional[str]
) -> Optional[str]:
    if value and value not in VALID_ON_ERROR_CHOICES:
        raise click.BadParameter(
            f"valid choices are: {', '.join(VALID_ON_ERROR_CHOICES)}"
        )
    return value


@click.group(name="workouts")
def workouts_cli() -> None:
    """Manage workouts."""
    pass


@workouts_cli.command("archive_uploads")
@click.option(
    "--max",
    "max_archives",
    type=int,
    required=True,
    help="Maximum number of workouts archive to process.",
)
@click.option(
    "--verbose",
    "-v",
    "verbose",
    is_flag=True,
    default=False,
    help="Enable verbose output log (default: disabled).",
)
def workouts_archive_uploads(max_archives: int, verbose: bool) -> None:
    """
    Process workouts archive uploads if queued tasks exist (progress = 0 and
    not aborted/errored).
    To use in case redis is not set.
    """
    with app.app_context():
        logger.setLevel(logging.DEBUG if verbose else logging.INFO)
        count = process_workouts_archives_uploads(max_archives, logger)
        logger.info(f"\nWorkouts archives processed: {count}.")


@workouts_cli.command("archive_upload")
@click.option(
    "--id",
    "task_short_id",
    type=str,
    required=True,
    help="Id of task to process",
)
@click.option(
    "--verbose",
    "-v",
    "verbose",
    is_flag=True,
    default=False,
    help="Enable verbose output log (default: disabled).",
)
def process_queued_archive_upload(task_short_id: str, verbose: bool) -> None:
    """
    Process given queued workouts archive upload.

    To use in case redis is not set.
    """
    with app.app_context():
        logger.setLevel(logging.DEBUG if verbose else logging.INFO)
        try:
            process_workouts_archive_upload(task_short_id, logger)
        except Exception as e:
            logger.error(str(e))
            sys.exit(1)
        logger.info("\nDone.")


def refresh_workouts_with_file(
    logger_: "Logger",
    sport_id: Optional[int] = None,
    new_sport_id: Optional[int] = None,
    date_from: Optional[datetime] = None,
    date_to: Optional[datetime] = None,
    per_page: int = 10,
    page: int = 1,
    order: str = "asc",
    user: Optional[str] = None,
    extension: Optional[str] = None,
    with_weather: bool = False,
    with_elevation: bool = False,
    on_file_error: Optional[str] = None,
    verbose: bool = False,
) -> None:
    if with_elevation:
        if (
            app.config["OPEN_ELEVATION_API_URL"]
            or app.config["VALHALLA_API_URL"]
        ):
            click.secho(
                "\nWarning: Open Elevation and/or Valhalla API "
                "are set.\n"
                "If users have enabled missing elevation processing, "
                "multiple calls will be made to the elevation services "
                "depending on the number of workouts to be refreshed, "
                "which may result in rate limit errors.\n"
                "Adjust the number of workouts to refresh with "
                "'--per-page' option if necessary "
                f"(current value: {per_page}).",
                fg="yellow",
                bold=True,
            )
            if not click.confirm("Do you want to continue?"):
                click.echo("Aborted!")
                sys.exit(0)
        else:
            click.secho(
                "\nWarning: '--with-elevation' is provided but "
                "no elevation API URL is set.\n",
                fg="yellow",
            )
    try:
        service = WorkoutsFromFileRefreshService(
            sport_id=sport_id,
            new_sport_id=new_sport_id,
            date_from=date_from,
            date_to=date_to,
            per_page=per_page,
            page=page,
            order=order,
            user=user,
            extension=extension,
            with_weather=with_weather,
            verbose=verbose,
            with_elevation=with_elevation,
            on_file_error=on_file_error,
            logger=logger_,
        )
        service.refresh()
    except Exception as e:
        logger.error(str(e))
        sys.exit(1)


def refresh_workouts_without_file(
    logger_: "Logger",
    sport_id: Optional[int] = None,
    new_sport_id: Optional[int] = None,
    date_from: Optional[datetime] = None,
    date_to: Optional[datetime] = None,
    per_page: int = 10,
    page: int = 1,
    order: str = "asc",
    user: Optional[str] = None,
    verbose: bool = False,
) -> None:
    try:
        service = WorkoutsWithoutFileRefreshService(
            logger=logger_,
            sport_id=sport_id,
            new_sport_id=new_sport_id,
            date_from=date_from,
            date_to=date_to,
            per_page=per_page,
            page=page,
            order=order,
            user=user,
            verbose=verbose,
        )
        service.refresh()
    except Exception as e:
        logger.error(str(e))
        sys.exit(1)


@workouts_cli.command("refresh")
@click.option(
    "--sport-id",
    help="sport id",
    type=int,
    callback=validate_sport_id,
)
@click.option(
    "--new-sport-id",
    help=(
        "sport id to which the workouts will be associated "
        "(must be provided with '--sport-id')"
    ),
    type=int,
    callback=validate_sport_id,
)
@click.option(
    "--from",
    "date_from",
    help="start date (format: %Y-%m-%d)",
    callback=validate_date,
)
@click.option(
    "--to",
    "date_to",
    help="end date (format: %Y-%m-%d)",
    callback=validate_date,
)
@click.option(
    "--per-page",
    help="number of workouts per page (default: 10)",
    type=int,
    callback=validate_number,
    default=10,
)
@click.option(
    "--page",
    help="page number (default: 1)",
    type=int,
    callback=validate_number,
    default=1,
)
@click.option(
    "--order",
    help="workout date order: 'asc' or 'desc' (default: 'asc')",
    type=str,
    default="asc",
    callback=validate_order,
)
@click.option(
    "--user",
    help="username of workouts owner",
    type=str,
    callback=validate_user,
)
@click.option(
    "--extension",
    help=(
        "workout file extension "
        f"(valid values are: {WORKOUT_VALID_EXTENSIONS})"
    ),
    type=str,
    callback=validate_extension,
)
@click.option(
    "--with-weather",
    help=(
        "enable weather data collection if weather provider is set and "
        "workout has no weather data. "
        "WARNING: depending on subscription, the rate limit can be reached, "
        "leading to errors and preventing weather data being collected during "
        "next uploads until the limit is reset (default: disabled)"
    ),
    is_flag=True,
    show_default=True,
    default=False,
)
@click.option(
    "--with-elevation",
    help=(
        "enable elevation update when elevation provider is set and "
        "some elevation data are missing. "
        "If disabled, existing elevation are not removed when elevation data "
        "are missing in the original file. "
        "WARNING: depending on subscription, the rate limit can be reached, "
        "leading to errors and preventing elevation data being collected "
        "during next uploads until the limit is reset (default: disabled)"
    ),
    is_flag=True,
    show_default=True,
    default=False,
)
@click.option(
    "--on-file-error",
    help=(
        "action to perform when workout file is not found. If not provided, "
        "an error will be raised. Valid actions are: 'remove-references' ("
        "all files references will be removed and workout preserved but not "
        "updated since no file found) and 'delete-workout'."
    ),
    type=str,
    callback=validate_on_file_error,
)
@click.option(
    "--without-file",
    help=(
        "allow to refresh workouts without a file and created before v1.1.0 "
        "to recalculate pace values (by default refresh command only "
        "refreshes workouts created with a file). "
        "When provided only workouts without file and without paces are "
        "refreshed (in this case '--extension', '--with-weather', "
        "'--with-elevation' and '--on-file-error' options are ignored). "
        "When not provided, only workouts with file are refreshed "
        "(default: disabled)"
    ),
    is_flag=True,
    show_default=True,
    default=False,
)
@click.option(
    "--verbose",
    "-v",
    "verbose",
    is_flag=True,
    default=False,
    help="enable verbose output log (default: disabled)",
)
def refresh_workouts(
    sport_id: Optional[int] = None,
    new_sport_id: Optional[int] = None,
    date_from: Optional[datetime] = None,
    date_to: Optional[datetime] = None,
    per_page: int = 10,
    page: int = 1,
    order: str = "asc",
    user: Optional[str] = None,
    extension: Optional[str] = None,
    with_weather: bool = False,
    with_elevation: bool = False,
    on_file_error: Optional[str] = None,
    without_file: bool = False,
    verbose: bool = False,
) -> None:
    """
    Refresh workouts
    """
    with app.app_context():
        logger.setLevel(logging.DEBUG if verbose else logging.INFO)

        if new_sport_id and not sport_id:
            raise click.BadOptionUsage(
                "--new-sport-id",
                "'--new-sport-id' must be provided with '--sport-id'",
            )

        if without_file:
            if (
                extension is not None
                or with_weather
                or with_elevation
                or on_file_error
            ):
                click.secho(
                    "\nWarning: when '--without-file' is provided, following "
                    "options are ignored: '--extension', '--with-weather', "
                    "'--with-elevation' and '--on-file-error'.\n",
                    fg="yellow",
                )

            refresh_workouts_without_file(
                logger_=logger,
                sport_id=sport_id,
                new_sport_id=new_sport_id,
                date_from=date_from,
                date_to=date_to,
                per_page=per_page,
                page=page,
                order=order,
                user=user,
                verbose=verbose,
            )
        else:
            refresh_workouts_with_file(
                logger_=logger,
                sport_id=sport_id,
                new_sport_id=new_sport_id,
                date_from=date_from,
                date_to=date_to,
                per_page=per_page,
                page=page,
                order=order,
                user=user,
                extension=extension,
                with_weather=with_weather,
                verbose=verbose,
                with_elevation=with_elevation,
                on_file_error=on_file_error,
            )

        logger.info("\nDone.")



"""add equipments visibility

Revision ID: ce68b3914ff7
Revises: 070bd31919c3
Create Date: 2025-02-02 10:29:28.718139

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ce68b3914ff7'
down_revision = '070bd31919c3'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('equipments', schema=None) as batch_op:
        batch_op.add_column(sa.Column('visibility', sa.Enum('PUBLIC', 'FOLLOWERS', 'PRIVATE', name='visibility_levels'), server_default='PRIVATE', nullable=False))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('equipments', schema=None) as batch_op:
        batch_op.drop_column('visibility')

    # ### end Alembic commands ###



from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Union
from uuid import UUID, uuid4

from sqlalchemy.dialects import postgresql
from sqlalchemy.engine.base import Connection
from sqlalchemy.event import listens_for
from sqlalchemy.orm import Mapped, Mapper, Session, mapped_column, relationship

from fittrackee import BaseModel, db
from fittrackee.comments.exceptions import CommentForbiddenException
from fittrackee.comments.models import Comment
from fittrackee.database import TZDateTime
from fittrackee.dates import aware_utc_now
from fittrackee.users.models import Notification, User
from fittrackee.users.roles import UserRole
from fittrackee.utils import encode_uuid
from fittrackee.workouts.exceptions import WorkoutForbiddenException
from fittrackee.workouts.models import Workout

from .exceptions import (
    InvalidReportActionAppealException,
    InvalidReportActionAppealUserException,
    InvalidReportActionException,
    InvalidReporterException,
    InvalidReportException,
    ReportActionAppealForbiddenException,
    ReportActionForbiddenException,
    ReportCommentForbiddenException,
    ReportForbiddenException,
)

REPORT_OBJECT_TYPES = [
    "comment",
    "user",
    "workout",
]
REPORT_ACTION_TYPES = [
    "report_reopening",
    "report_resolution",
]
USER_ACTION_TYPES = [
    "user_suspension",
    "user_unsuspension",
    "user_warning",
]
ALL_USER_ACTION_TYPES = [
    *USER_ACTION_TYPES,
    "user_warning_lifting",
]
COMMENT_ACTION_TYPES = [
    "comment_suspension",
    "comment_unsuspension",
]
WORKOUT_ACTION_TYPES = [
    "workout_suspension",
    "workout_unsuspension",
]
OBJECTS_ACTION_TYPES = (
    COMMENT_ACTION_TYPES + ALL_USER_ACTION_TYPES + WORKOUT_ACTION_TYPES
)
ALL_ACTION_TYPES = REPORT_ACTION_TYPES + OBJECTS_ACTION_TYPES


class Report(BaseModel):
    __tablename__ = "reports"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    created_at: Mapped[datetime] = mapped_column(
        TZDateTime, default=aware_utc_now
    )
    updated_at: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, nullable=True
    )
    resolved_at: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, nullable=True
    )
    reported_by: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("users.id", ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    reported_comment_id: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("comments.id", ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    reported_user_id: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("users.id", ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    reported_workout_id: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("workouts.id", ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    resolved_by: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("users.id", ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    resolved: Mapped[bool] = mapped_column(default=False, nullable=False)
    object_type: Mapped[str] = mapped_column(
        db.String(50), nullable=False, index=True
    )
    note: Mapped[str] = mapped_column(db.String(), nullable=False)

    reported_comment: Mapped["Comment"] = relationship("Comment", lazy=True)
    reported_user: Mapped["User"] = relationship(
        "User", primaryjoin=reported_user_id == User.id
    )
    reported_workout: Mapped["Workout"] = relationship("Workout", lazy=True)
    reporter: Mapped["User"] = relationship(
        "User", primaryjoin=reported_by == User.id
    )
    resolver: Mapped["User"] = relationship(
        "User", primaryjoin=resolved_by == User.id
    )
    comments: Mapped[List["ReportComment"]] = relationship(
        "ReportComment", back_populates="report"
    )
    report_actions: Mapped[List["ReportAction"]] = relationship(
        "ReportAction",
        lazy=True,
        back_populates="report",
        order_by="ReportAction.created_at.asc()",
    )

    @property
    def reported_object(self) -> Union[Comment, None, User, Workout]:
        # util method, used by tests
        if self.object_type == "comment":
            return self.reported_comment
        if self.object_type == "user":
            return self.reported_user
        if self.object_type == "workout":
            return self.reported_workout
        return None

    @property
    def is_reported_user_warned(self) -> bool:
        return (
            ReportAction.query.filter_by(
                action_type="user_warning",
                report_id=self.id,
                user_id=self.reported_user_id,
            ).first()
            is not None
        )

    def __init__(
        self,
        note: str,
        reported_by: int,
        reported_object: Union[Comment, User, Workout],
        created_at: Optional[datetime] = None,
    ):
        object_type = reported_object.__class__.__name__.lower()
        if object_type not in REPORT_OBJECT_TYPES:
            raise InvalidReportException()
        user_id = (
            reported_object.id
            if isinstance(reported_object, User)
            else reported_object.user_id
        )
        if user_id == reported_by:
            raise InvalidReporterException()

        self.created_at = (
            created_at if created_at else datetime.now(timezone.utc)
        )
        self.note = note
        self.object_type = object_type
        self.reported_by = reported_by
        self.reported_comment_id = (
            reported_object.id
            if isinstance(reported_object, Comment)
            else None
        )
        self.reported_user_id = user_id
        self.reported_workout_id = (
            reported_object.id
            if isinstance(reported_object, Workout)
            else None
        )
        self.resolved = False

    def serialize(self, current_user: User, full: bool = False) -> Dict:
        if (
            not current_user.has_moderator_rights
            and self.reported_by != current_user.id
        ):
            raise ReportForbiddenException()

        try:
            reported_comment = (
                self.reported_comment.serialize(current_user, for_report=True)
                if self.reported_comment_id
                else None
            )
        except CommentForbiddenException:
            reported_comment = {"id": "_COMMENT_UNAVAILABLE_"}

        try:
            reported_workout = (
                self.reported_workout.serialize(
                    user=current_user, for_report=True
                )
                if self.reported_workout
                else None
            )
        except WorkoutForbiddenException:
            reported_workout = {"id": "_WORKOUT_UNAVAILABLE_"}

        report = {
            "created_at": self.created_at,
            "id": self.id,
            "is_reported_user_warned": self.is_reported_user_warned,
            "note": self.note,
            "object_type": self.object_type,
            "reported_by": (
                self.reporter.serialize(current_user=current_user)
                if self.reported_by
                else None
            ),
            "reported_comment": reported_comment,
            "reported_user": (
                self.reported_user.serialize(current_user=current_user)
                if self.reported_user_id
                else None
            ),
            "reported_workout": reported_workout,
            "resolved": self.resolved,
            "resolved_at": self.resolved_at,
        }
        if current_user.has_moderator_rights:
            if full:
                report["report_actions"] = [
                    action.serialize(current_user, full=False)
                    for action in self.report_actions
                ]
                report["comments"] = [
                    comment.serialize(current_user)
                    for comment in self.comments
                ]
            report["resolved_by"] = (
                None
                if self.resolved_by is None
                else self.resolver.serialize(current_user=current_user)
            )
            report["updated_at"] = self.updated_at
        return report


@listens_for(Report, "after_insert")
def on_report_insert(
    mapper: Mapper, connection: Connection, new_report: Report
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        if not new_report.reported_by:
            return

        from fittrackee.users.models import Notification, User

        for admin in User.query.filter(
            User.role >= UserRole.MODERATOR.value,
            User.id != new_report.reported_by,
            User.is_active == True,  # noqa
        ).all():
            notification = Notification(
                from_user_id=new_report.reported_by,
                to_user_id=admin.id,
                created_at=new_report.created_at,
                event_type="report",
                event_object_id=new_report.id,
            )
            session.add(notification)


class ReportComment(BaseModel):
    __tablename__ = "report_comments"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    created_at: Mapped[datetime] = mapped_column(
        TZDateTime, default=aware_utc_now
    )
    report_id: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("reports.id", ondelete="CASCADE"),
        index=True,
        nullable=True,
    )
    user_id: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("users.id", ondelete="CASCADE"),
        index=True,
        nullable=True,
    )
    comment: Mapped[str] = mapped_column(db.String(), nullable=False)

    user: Mapped["User"] = relationship("User")
    report: Mapped["Report"] = relationship(
        "Report",
        lazy="select",
        single_parent=True,
    )

    def __init__(
        self,
        report_id: int,
        user_id: int,
        comment: str,
        created_at: Optional[datetime] = None,
    ):
        self.created_at = (
            created_at if created_at else datetime.now(timezone.utc)
        )
        self.comment = comment
        self.report_id = report_id
        self.user_id = user_id

    def serialize(self, current_user: User) -> Dict:
        if not current_user.has_moderator_rights:
            raise ReportCommentForbiddenException()
        return {
            "created_at": self.created_at,
            "comment": self.comment,
            "id": self.id,
            "report_id": self.report_id,
            "user": self.user.serialize(current_user=current_user),
        }


class ReportAction(BaseModel):
    __tablename__ = "report_actions"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    uuid: Mapped[UUID] = mapped_column(
        postgresql.UUID(as_uuid=True),
        default=uuid4,
        unique=True,
        nullable=False,
    )
    created_at: Mapped[datetime] = mapped_column(
        TZDateTime, default=aware_utc_now, index=True
    )
    moderator_id: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("users.id", ondelete="SET NULL"),
        index=True,
        nullable=True,  # to keep log if an moderator is deleted
    )
    report_id: Mapped[int] = mapped_column(
        db.ForeignKey("reports.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    user_id: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("users.id", ondelete="CASCADE"),
        index=True,
        nullable=True,
    )
    workout_id: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("workouts.id", ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    comment_id: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("comments.id", ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    action_type: Mapped[str] = mapped_column(db.String(50), nullable=False)
    reason: Mapped[Optional[str]] = mapped_column(db.String(), nullable=True)

    moderator: Mapped["User"] = relationship(
        "User",
        primaryjoin=moderator_id == User.id,
        lazy="joined",
        single_parent=True,
    )
    user: Mapped["User"] = relationship(
        "User",
        primaryjoin=user_id == User.id,
        lazy="joined",
        single_parent=True,
    )
    appeal: Mapped["ReportActionAppeal"] = relationship(
        "ReportActionAppeal",
        uselist=False,
        back_populates="action",
    )
    comment: Mapped["Comment"] = relationship("Comment", lazy=True)
    workout: Mapped["Workout"] = relationship("Workout", lazy=True)
    report: Mapped["Report"] = relationship(
        "Report", lazy="select", single_parent=True
    )

    def __init__(
        self,
        action_type: str,
        moderator_id: int,
        report_id: int,
        *,
        user_id: Optional[int] = None,
        comment_id: Optional[int] = None,
        workout_id: Optional[int] = None,
        reason: Optional[str] = None,
        created_at: Optional[datetime] = None,
    ):
        if action_type not in ALL_ACTION_TYPES:
            raise InvalidReportActionException()
        if action_type in OBJECTS_ACTION_TYPES and not user_id:
            raise InvalidReportActionException()
        if action_type in WORKOUT_ACTION_TYPES and not workout_id:
            raise InvalidReportActionException()
        if action_type in COMMENT_ACTION_TYPES and not comment_id:
            raise InvalidReportActionException()

        self.action_type = action_type
        self.moderator_id = moderator_id
        self.created_at = (
            created_at if created_at else datetime.now(timezone.utc)
        )
        self.comment_id = (
            comment_id if action_type in COMMENT_ACTION_TYPES else None
        )
        self.reason = reason
        self.report_id = report_id
        self.user_id = user_id if action_type in OBJECTS_ACTION_TYPES else None
        self.workout_id = (
            workout_id if action_type in WORKOUT_ACTION_TYPES else None
        )

    @property
    def short_id(self) -> str:
        return encode_uuid(self.uuid)

    def serialize(self, current_user: User, full: bool = True) -> Dict:
        if (
            not current_user.has_moderator_rights
            and current_user.id != self.user_id
        ):
            raise ReportActionForbiddenException()
        action: Dict[str, Any] = {
            "action_type": self.action_type,
            "appeal": (
                self.appeal.serialize(current_user) if self.appeal else None
            ),
            "created_at": self.created_at,
            "id": self.short_id,
            "reason": self.reason,
        }

        if current_user.has_moderator_rights:
            action["report_id"] = self.report_id
            action["moderator"] = self.moderator.serialize(
                current_user=current_user
            )
            action["user"] = (
                self.user.serialize(current_user=current_user)
                if self.user
                else None
            )
        if not full:
            return action

        if current_user.has_moderator_rights:
            action = {
                **action,
                "comment": (
                    self.comment.serialize(user=current_user, for_report=True)
                    if self.comment_id
                    else None
                ),
                "workout": (
                    self.workout.serialize(user=current_user, for_report=True)
                    if self.workout_id
                    else None
                ),
            }
        else:
            if self.report.object_type == "comment":
                action["comment"] = (
                    self.comment.serialize(user=current_user)
                    if self.comment_id
                    else None
                )
            if self.report.object_type == "workout":
                action["workout"] = (
                    self.workout.serialize(user=current_user)
                    if self.workout_id
                    else None
                )

        return action


class ReportActionAppeal(BaseModel):
    __tablename__ = "report_action_appeals"
    __table_args__ = (
        db.UniqueConstraint(
            "action_id", "user_id", name="action_id_user_id_unique"
        ),
    )

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    uuid: Mapped[UUID] = mapped_column(
        postgresql.UUID(as_uuid=True),
        default=uuid4,
        unique=True,
        nullable=False,
    )
    action_id: Mapped[int] = mapped_column(
        db.ForeignKey("report_actions.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    moderator_id: Mapped[Optional[int]] = mapped_column(
        db.ForeignKey("users.id", ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    created_at: Mapped[datetime] = mapped_column(
        TZDateTime, default=aware_utc_now, nullable=False
    )
    updated_at: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, nullable=True
    )
    approved: Mapped[Optional[bool]] = mapped_column(nullable=True)
    text: Mapped[str] = mapped_column(db.String(), nullable=False)
    reason: Mapped[Optional[str]] = mapped_column(db.String(), nullable=True)

    moderator: Mapped["User"] = relationship(
        "User",
        primaryjoin=moderator_id == User.id,
        lazy="joined",
        single_parent=True,
    )
    user: Mapped["User"] = relationship(
        "User",
        primaryjoin=user_id == User.id,
        lazy="joined",
        single_parent=True,
    )
    action: Mapped["ReportAction"] = relationship(
        "ReportAction", lazy="joined", single_parent=True
    )

    def __init__(
        self,
        action_id: int,
        user_id: int,
        text: str,
        created_at: Optional[datetime] = None,
    ):
        action = ReportAction.query.filter_by(id=action_id).one()
        if action.action_type not in [
            "comment_suspension",
            "user_suspension",
            "user_warning",
            "workout_suspension",
        ]:
            raise InvalidReportActionAppealException()
        if action.user_id != user_id:
            raise InvalidReportActionAppealUserException()
        self.action_id = action_id
        self.created_at = (
            created_at if created_at else datetime.now(timezone.utc)
        )
        self.text = text
        self.user_id = user_id

    @property
    def short_id(self) -> str:
        return encode_uuid(self.uuid)

    def serialize(self, current_user: User) -> Dict:
        if (
            not current_user.has_moderator_rights
            and current_user.id != self.user_id
        ):
            raise ReportActionAppealForbiddenException()
        appeal: Dict[str, Any] = {
            "approved": self.approved,
            "created_at": self.created_at,
            "id": self.short_id,
            "reason": self.reason,
            "text": self.text,
            "updated_at": self.updated_at,
        }
        if current_user.has_moderator_rights:
            appeal["moderator"] = (
                self.moderator.serialize(current_user=current_user)
                if self.moderator
                else None
            )
            appeal["user"] = self.user.serialize(current_user=current_user)
        return appeal


@listens_for(ReportAction, "after_insert")
def on_report_action_insert(
    mapper: Mapper, connection: Connection, new_action: ReportAction
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        if not new_action.moderator_id or not new_action.user_id:
            return

        if (
            new_action.action_type
            in COMMENT_ACTION_TYPES
            + WORKOUT_ACTION_TYPES
            + ["user_warning", "user_warning_lifting"]
        ):
            notification = Notification(
                from_user_id=new_action.moderator_id,
                to_user_id=new_action.user_id,
                created_at=new_action.created_at,
                event_type=new_action.action_type,
                event_object_id=new_action.id,
            )
            session.add(notification)


@listens_for(ReportActionAppeal, "after_insert")
def on_report_action_appeal_insert(
    mapper: Mapper, connection: Connection, new_appeal: ReportActionAppeal
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        from fittrackee.users.models import Notification, User

        report_action = ReportAction.query.filter_by(
            id=new_appeal.action_id
        ).first()
        if report_action:
            for admin in User.query.filter(
                User.role >= UserRole.MODERATOR.value,
                User.id != new_appeal.user_id,
                User.is_active == True,  # noqa
            ).all():
                notification = Notification(
                    from_user_id=new_appeal.user_id,
                    to_user_id=admin.id,
                    created_at=new_appeal.created_at,
                    event_type=(
                        "user_warning_appeal"
                        if report_action.action_type == "user_warning"
                        else "suspension_appeal"
                    ),
                    event_object_id=new_appeal.id,
                )
                session.add(notification)



"""add application config in database

Revision ID: 8a0aad4c838c
Revises: 1345afe3b11d
Create Date: 2019-11-13 13:14:20.147296

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '8a0aad4c838c'
down_revision = '1345afe3b11d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('app_config',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('max_users', sa.Integer(), nullable=False),
    sa.Column('gpx_limit_import', sa.Integer(), nullable=False),
    sa.Column('max_single_file_size', sa.Integer(), nullable=False),
    sa.Column('max_zip_file_size', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('app_config')
    # ### end Alembic commands ###



import zipfile
from typing import IO, TYPE_CHECKING

from ...exceptions import WorkoutFileException
from .workout_kml_service import WorkoutKmlService

if TYPE_CHECKING:
    import gpxpy.gpx


class WorkoutKmzService(WorkoutKmlService):
    @classmethod
    def parse_file(
        cls,
        workout_file: IO[bytes],
        segments_creation_event: str,
    ) -> "gpxpy.gpx.GPX":
        """
        Only kmz files with Tracks are supported.
        Notes:
        - for now kmz with photos are not supported
        - segments_creation_event is not used (only for .fit files)

        Tested with files generated with OpenTracks.
        """

        # file is already open with Zipfile (in case of workout refresh)
        if not zipfile.is_zipfile(workout_file):
            workout_file.seek(0)
            return super().parse_file(workout_file, segments_creation_event)

        with zipfile.ZipFile(workout_file, "r") as kmz_ref:
            try:
                kml_content = kmz_ref.open("doc.kml")
            except KeyError as e:
                raise WorkoutFileException(
                    "error", "error when parsing kmz file"
                ) from e
        return super().parse_file(kml_content, segments_creation_event)



from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Optional

import humanize
import pytz
from babel.dates import format_datetime

from fittrackee.languages import LANGUAGES_DATE_STRING
from fittrackee.users.constants import USER_DATE_FORMAT, USER_TIMEZONE
from fittrackee.users.utils.language import get_language

if TYPE_CHECKING:
    from fittrackee.users.models import User


def get_date_string_for_user(date_to_format: datetime, user: "User") -> str:
    """
    Note: date_to_format is a timezone-aware datetime in UTC
    """
    user_language = get_language(user.language)
    user_timezone = user.timezone if user.timezone else USER_TIMEZONE
    user_date_format = (
        user.date_format if user.date_format else USER_DATE_FORMAT
    )

    date_format = (
        LANGUAGES_DATE_STRING[user_language]
        if user_date_format == "date_string"
        else user_date_format
    )
    return format_datetime(
        date_to_format.astimezone(pytz.timezone(user_timezone)),
        format=f"{date_format} - HH:mm:ss",
        locale=user_language,
    )


def get_readable_duration(duration: int, locale: Optional[str] = None) -> str:
    """
    Return readable and localized duration from duration in seconds
    """
    if locale is None:
        locale = "en"
    if locale != "en":
        try:
            _t = humanize.i18n.activate(locale)
        except FileNotFoundError:
            locale = "en"
    readable_duration = humanize.naturaldelta(timedelta(seconds=duration))
    if locale != "en":
        humanize.i18n.deactivate()
    return readable_duration


def aware_utc_now() -> datetime:
    return datetime.now(timezone.utc)


def get_datetime_in_utc(
    date_string: str, date_format: str = "%Y-%m-%d"
) -> datetime:
    return datetime.strptime(date_string, date_format).replace(
        tzinfo=timezone.utc
    )



import os
from datetime import datetime
from typing import Dict, Optional

from flask import current_app
from sqlalchemy.engine.base import Connection
from sqlalchemy.event import listens_for
from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import text

from fittrackee import BaseModel, db
from fittrackee.database import TZDateTime
from fittrackee.users.models import User


class AppConfig(BaseModel):
    __tablename__ = "app_config"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    max_users: Mapped[int] = mapped_column(default=0, nullable=False)
    file_sync_limit_import: Mapped[int] = mapped_column(
        default=10, nullable=False
    )
    max_single_file_size: Mapped[int] = mapped_column(
        default=1048576, nullable=False
    )
    max_zip_file_size: Mapped[int] = mapped_column(
        default=10485760, nullable=False
    )
    admin_contact: Mapped[Optional[str]] = mapped_column(
        db.String(255), nullable=True
    )
    privacy_policy_date: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, nullable=True
    )
    privacy_policy: Mapped[Optional[str]] = mapped_column(
        db.Text, nullable=True
    )
    about: Mapped[Optional[str]] = mapped_column(db.Text, nullable=True)
    stats_workouts_limit: Mapped[int] = mapped_column(
        default=10000, nullable=False
    )
    file_limit_import: Mapped[int] = mapped_column(default=10, nullable=False)
    global_map_workouts_limit: Mapped[int] = mapped_column(
        server_default="10000", nullable=False
    )

    @property
    def is_registration_enabled(self) -> bool:
        result = db.session.execute(text("SELECT COUNT(*) FROM users;"))
        nb_users = result.one()[0]
        return self.max_users == 0 or nb_users < self.max_users

    @property
    def map_attribution(self) -> str:
        return current_app.config["TILE_SERVER"]["ATTRIBUTION"]

    @property
    def elevation_services(self) -> Dict:
        return {
            "open_elevation": (
                current_app.config["OPEN_ELEVATION_API_URL"] != ""
            ),
            "valhalla": current_app.config["VALHALLA_API_URL"] != "",
        }

    def serialize(self) -> Dict:
        weather_provider = os.getenv("WEATHER_API_PROVIDER", "").lower()
        return {
            "about": self.about,
            "admin_contact": self.admin_contact,
            "elevation_services": self.elevation_services,
            "file_limit_import": self.file_limit_import,
            "file_sync_limit_import": self.file_sync_limit_import,
            "is_email_sending_enabled": current_app.config["CAN_SEND_EMAILS"],
            "is_registration_enabled": self.is_registration_enabled,
            "global_map_workouts_limit": self.global_map_workouts_limit,
            "max_single_file_size": self.max_single_file_size,
            "max_zip_file_size": self.max_zip_file_size,
            "max_users": self.max_users,
            "map_attribution": self.map_attribution,
            "privacy_policy": self.privacy_policy,
            "privacy_policy_date": (
                self.privacy_policy_date
                if self.privacy_policy
                else current_app.config["DEFAULT_PRIVACY_POLICY_DATA"]
            ),
            "stats_workouts_limit": self.stats_workouts_limit,
            "version": current_app.config["VERSION"],
            "weather_provider": (
                weather_provider
                if weather_provider in ["visualcrossing"]
                else None
            ),
        }


def update_app_config() -> None:
    config = AppConfig.query.first()
    if config:
        current_app.config["is_registration_enabled"] = (
            config.is_registration_enabled
        )


@listens_for(User, "after_insert")
def on_user_insert(mapper: Mapper, connection: Connection, user: User) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        update_app_config()


@listens_for(User, "after_delete")
def on_user_delete(
    mapper: Mapper, connection: Connection, old_user: User
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        update_app_config()



"""add OAuth 2.0 and blacklisted tokens

Revision ID: 84d840ce853b
Revises: 5e3a3a31c432
Create Date: 2022-05-27 10:54:02.284543

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '84d840ce853b'
down_revision = 'cd0e6cf83207'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('oauth2_client',
    sa.Column('client_id', sa.String(length=48), nullable=True),
    sa.Column('client_secret', sa.String(length=120), nullable=True),
    sa.Column('client_id_issued_at', sa.Integer(), nullable=False),
    sa.Column('client_secret_expires_at', sa.Integer(), nullable=False),
    sa.Column('client_metadata', sa.Text(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_oauth2_client_client_id'), 'oauth2_client', ['client_id'], unique=False)
    op.create_index(op.f('ix_oauth2_client_user_id'), 'oauth2_client', ['user_id'], unique=False)
    op.create_table('oauth2_code',
    sa.Column('code', sa.String(length=120), nullable=False),
    sa.Column('client_id', sa.String(length=48), nullable=True),
    sa.Column('redirect_uri', sa.Text(), nullable=True),
    sa.Column('response_type', sa.Text(), nullable=True),
    sa.Column('scope', sa.Text(), nullable=True),
    sa.Column('nonce', sa.Text(), nullable=True),
    sa.Column('auth_time', sa.Integer(), nullable=False),
    sa.Column('code_challenge', sa.Text(), nullable=True),
    sa.Column('code_challenge_method', sa.String(length=48), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('code')
    )
    op.create_index('ix_oauth2_code_client_id', 'oauth2_code', ['client_id'], unique=False)
    op.create_index(op.f('ix_oauth2_code_user_id'), 'oauth2_code', ['user_id'], unique=False)
    op.create_table('oauth2_token',
    sa.Column('client_id', sa.String(length=48), nullable=True),
    sa.Column('token_type', sa.String(length=40), nullable=True),
    sa.Column('access_token', sa.String(length=255), nullable=False),
    sa.Column('refresh_token', sa.String(length=255), nullable=True),
    sa.Column('scope', sa.Text(), nullable=True),
    sa.Column('issued_at', sa.Integer(), nullable=False),
    sa.Column('access_token_revoked_at', sa.Integer(), nullable=False),
    sa.Column('refresh_token_revoked_at', sa.Integer(), nullable=False),
    sa.Column('expires_in', sa.Integer(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('access_token')
    )
    op.create_index(op.f('ix_oauth2_token_refresh_token'), 'oauth2_token', ['refresh_token'], unique=False)
    op.create_index(op.f('ix_oauth2_token_user_id'), 'oauth2_token', ['user_id'], unique=False)

    op.create_table('blacklisted_tokens',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('token', sa.String(length=500), nullable=False),
    sa.Column('expired_at', sa.Integer(), nullable=False),
    sa.Column('blacklisted_on', sa.DateTime(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('token')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('blacklisted_tokens')
    op.drop_index(op.f('ix_oauth2_token_user_id'), table_name='oauth2_token')
    op.drop_index(op.f('ix_oauth2_token_refresh_token'), table_name='oauth2_token')
    op.drop_table('oauth2_token')
    op.drop_index(op.f('ix_oauth2_code_user_id'), table_name='oauth2_code')
    op.drop_index('ix_oauth2_code_client_id', table_name='oauth2_code')
    op.drop_table('oauth2_code')
    op.drop_index(op.f('ix_oauth2_client_user_id'), table_name='oauth2_client')
    op.drop_index(op.f('ix_oauth2_client_client_id'), table_name='oauth2_client')
    op.drop_table('oauth2_client')
    # ### end Alembic commands ###



from typing import Optional

from flask import current_app

from fittrackee.users.constants import USER_LANGUAGE


def get_language(language: Optional[str]) -> str:
    # Note: some users may not have language preferences set
    if not language or language not in current_app.config["LANGUAGES"]:
        language = USER_LANGUAGE
    return language



import zipfile
from datetime import datetime
from typing import IO, TYPE_CHECKING, Optional, Union

from fittrackee import appLog, db
from fittrackee.files import get_absolute_file_path
from fittrackee.users.models import User, UserSportPreference
from fittrackee.workouts.models import Workout, WorkoutSegment

from ..exceptions import (
    WorkoutExceedingValueException,
    WorkoutException,
    WorkoutFileException,
    WorkoutRefreshException,
)
from .abstract_workouts_refresh_service import AbstractWorkoutsRefreshService
from .mixins import WorkoutFileMixin
from .workout_from_file.services import WORKOUT_FROM_FILE_SERVICES

if TYPE_CHECKING:
    from logging import Logger

    from fittrackee.constants import ElevationDataSource
    from fittrackee.workouts.models import Sport


class WorkoutFromFileRefreshService(WorkoutFileMixin):
    """
    recalculate values ang regenerate geometry and points
    """

    def __init__(
        self,
        workout: "Workout",
        update_weather: bool = False,
        get_elevation_on_refresh: bool = True,
        change_elevation_source: Optional["ElevationDataSource"] = None,
        on_file_error: Optional[str] = None,
        logger: Optional["Logger"] = None,
    ):
        if not workout.original_file:
            raise WorkoutRefreshException(
                "error", "workout without original file"
            )

        self.original_file: str = workout.original_file
        self.workout: "Workout" = workout
        self.user: "User" = workout.user
        self.sport: "Sport" = workout.sport
        self.sport_preferences = UserSportPreference.query.filter_by(
            user_id=self.user.id, sport_id=self.sport.id
        ).first()
        self.stopped_speed_threshold = (
            self.sport.stopped_speed_threshold
            if self.sport_preferences is None
            else self.sport_preferences.stopped_speed_threshold
        )
        self.update_weather = update_weather
        self.get_elevation_on_refresh = get_elevation_on_refresh
        self.change_elevation_source = change_elevation_source
        self.on_file_error = on_file_error
        self.logger = logger

    def get_file_content(self, file_extension: str) -> Union[bytes, IO[bytes]]:
        try:
            file_path = get_absolute_file_path(self.original_file)

            if file_extension == "kmz":
                with zipfile.ZipFile(file_path, "r") as kmz_ref:
                    return kmz_ref.open("doc.kml")

            with open(file_path, "rb") as f:
                file_content = f.read()
                return file_content
        except Exception:
            raise WorkoutFileException(
                "error", "error when opening original file"
            ) from None

    def _log_message(self, message: str) -> None:
        if self.logger:
            self.logger.info(message)

    def refresh(self) -> Optional["Workout"]:
        file_extension = self._get_extension(self.original_file)
        if not self._is_valid_workout_file_extension(file_extension):
            raise WorkoutRefreshException("error", "invalid file extension")

        try:
            file_content = self.get_file_content(file_extension)
        except WorkoutFileException as e:
            if self.on_file_error == "delete-workout":
                workout_id = self.workout.short_id
                user_name = self.workout.user.username
                db.session.delete(self.workout)
                db.session.commit()
                self._log_message(
                    f"No file found for workout '{workout_id}' (user: "
                    f"{user_name}), workout deleted."
                )
                return None
            if self.on_file_error == "remove-references":
                WorkoutSegment.query.filter_by(
                    workout_id=self.workout.id
                ).delete()
                self.workout.original_file = None
                self.workout.map_id = None
                self.workout.map = None
                db.session.commit()
                self._log_message(
                    f"No file found for workout '{self.workout.short_id}' "
                    f"(user: {self.workout.user.username}), segments deleted "
                    "and files references removed."
                )
                return self.workout
            raise e

        workout_service = WORKOUT_FROM_FILE_SERVICES[file_extension](
            auth_user=self.user,
            workout_file=file_content,  # type: ignore[arg-type]
            workout=self.workout,
            sport=self.sport,
            stopped_speed_threshold=self.stopped_speed_threshold,
            get_weather=self.update_weather,
            get_elevation_on_refresh=self.get_elevation_on_refresh,
            change_elevation_source=self.change_elevation_source,
        )

        # extract and calculate data from provided file
        # note: workout date is not updated
        try:
            workout_service.process_workout()
        except (WorkoutExceedingValueException, WorkoutFileException) as e:
            appLog.exception(f"workout exception: {e!s}")
            db.session.rollback()
            raise e
        except Exception as e:
            appLog.exception(f"exception: {e!s}")
            db.session.rollback()
            raise WorkoutException(
                "error", "error when processing workout"
            ) from e

        db.session.commit()
        db.session.refresh(self.workout)
        return self.workout


class WorkoutsFromFileRefreshService(AbstractWorkoutsRefreshService):
    """
    All checks on parameters are made by the CLI command.
    """

    def __init__(
        self,
        logger: "Logger",
        per_page: Optional[int] = 10,
        page: Optional[int] = 1,
        order: Optional[str] = "asc",
        user: Optional[str] = None,
        extension: Optional[str] = None,
        sport_id: Optional[int] = None,
        new_sport_id: Optional[int] = None,
        date_from: Optional[datetime] = None,
        date_to: Optional[datetime] = None,
        with_weather: bool = False,
        with_elevation: bool = False,
        on_file_error: Optional[str] = None,
        verbose: bool = False,
    ) -> None:
        super().__init__(
            logger,
            per_page,
            page,
            order,
            user,
            sport_id,
            new_sport_id,
            date_from,
            date_to,
            verbose,
        )
        self.extension: Optional[str] = extension if extension else None
        self.with_weather: bool = with_weather
        self.with_elevation: bool = with_elevation
        self.on_file_error: Optional[str] = on_file_error

    def refresh(self) -> int:
        filters = [Workout.original_file != None]  # noqa
        if self.extension:
            filters.extend([Workout.original_file.like(f"%{self.extension}")])

        workouts_to_refresh, total = self._get_workouts(filters)
        if not total:
            return 0

        updated_workouts = 0
        deleted_workouts = 0
        for index, workout in enumerate(workouts_to_refresh, start=1):
            self._log_if_verbose(f"Refreshing workout {index}/{total}...")

            try:
                if self.new_sport_id:
                    workout.sport_id = self.new_sport_id
                    db.session.flush()
                    db.session.refresh(workout)
                service = WorkoutFromFileRefreshService(
                    workout,
                    update_weather=self.with_weather,
                    get_elevation_on_refresh=self.with_elevation,
                    on_file_error=self.on_file_error,
                    logger=self.logger,
                )
                refreshed_workout = service.refresh()
                if refreshed_workout:
                    updated_workouts += 1
                else:
                    deleted_workouts += 1
            except Exception as e:
                self.logger.error(
                    (
                        f"Error when refreshing workout '{workout.short_id}' "
                        f"(user: {workout.user.username}): {e}"
                    )
                )
                continue

        if total:
            deleted_workouts_count = (
                f"- deleted workouts: {deleted_workouts}\n"
                if deleted_workouts
                else ""
            )
            errored_workouts = total - (updated_workouts + deleted_workouts)
            self._log_if_verbose(
                "\nRefresh done:\n"
                f"- updated workouts: {updated_workouts}\n"
                f"{deleted_workouts_count}"
                f"- errored workouts: {errored_workouts}"
            )

        return updated_workouts



"""add ascent record

Revision ID: cd0e6cf83207
Revises: 5e3a3a31c432
Create Date: 2022-03-22 20:21:13.661883

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'cd0e6cf83207'
down_revision = '5e3a3a31c432'
branch_labels = None
depends_on = None


def upgrade():
    # workaround for PostgreSQL<12
    # (can not use ALTER TYPE to add values in migrations)
    op.execute("ALTER TYPE record_types RENAME TO record_types_old")
    op.execute("CREATE TYPE record_types AS ENUM('AS', 'FD', 'LD', 'MS', 'HA')")
    op.execute(
        """
        ALTER TABLE records ALTER COLUMN record_type TYPE record_types
        USING record_type::text::record_types
    """
    )
    op.execute("DROP TYPE record_types_old")

    op.add_column(
        'users', sa.Column('display_ascent', sa.Boolean(), nullable=True)
    )
    op.execute("UPDATE users SET display_ascent = true")
    op.alter_column('users', 'display_ascent', nullable=False)


def downgrade():
    op.drop_column('users', 'display_ascent')

    op.execute("DELETE FROM records WHERE record_type = 'HA';")
    op.execute("ALTER TYPE record_types RENAME TO record_types_old")
    op.execute("CREATE TYPE record_types AS ENUM('AS', 'FD', 'LD', 'MS')")
    op.execute(
        """
        ALTER TABLE records ALTER COLUMN record_type TYPE record_types
        USING record_type::text::record_types
    """
    )
    op.execute("DROP TYPE record_types_old")



"""add ave and max power

Revision ID: 45fd75e5f16f
Revises: 98688d3d293b
Create Date: 2025-07-19 23:26:26.509887

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '45fd75e5f16f'
down_revision = '98688d3d293b'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workout_segments', schema=None) as batch_op:
        batch_op.add_column(sa.Column('max_power', sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column('ave_power', sa.Integer(), nullable=True))

    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.add_column(sa.Column('max_power', sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column('ave_power', sa.Integer(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.drop_column('ave_power')
        batch_op.drop_column('max_power')

    with op.batch_alter_table('workout_segments', schema=None) as batch_op:
        batch_op.drop_column('ave_power')
        batch_op.drop_column('max_power')

    # ### end Alembic commands ###
